"""ACE. Option"""
from __future__ import annotations

from functools import total_ordering

from cisco_acl import helpers as h
from cisco_acl.base import Base
from cisco_acl.types_ import LStr, DAny

LOGS = ("log", "log-input")


@total_ordering
class Option(Base):
    """ACE. Option"""

    def __init__(self, line: str = "", **kwargs):
        """ACE. Option
        :param line: Option line
        :type line: str

        :param platform: Platform: "ios", "nxos" (default "ios")
        :type platform: str

        Helpers
        :param note: Object description
        :type note: Any

        :example:
            option = Option("ack dscp ef log")
            result:
                option.line == "ack log"
                option.flags == {"ack", "dscp", "ef"}
                option.logs == {"log"}
        """
        self._line: str = ""
        self._flags: LStr = []
        self._logs: LStr = []
        super().__init__(**kwargs)  # platform, note
        self.line = line

    # ========================== redefined ===========================

    def __hash__(self) -> int:
        return tuple(self._flags).__hash__()

    def __eq__(self, other) -> bool:
        """== equality"""
        return self.__hash__() == other.__hash__()

    def __lt__(self, other) -> bool:
        """< less than"""
        if self.__class__ == other.__class__:
            return self._flags < self._flags
        return False

    # =========================== property ===========================

    @property
    def flags(self) -> LStr:
        """ACE Option flags, items that related to packet forwarding"""
        return self._flags

    @property
    def line(self) -> str:
        """ACE Option items as **str"""
        return self._line

    @line.setter
    def line(self, line: str) -> None:
        line = h.init_line(line)
        self._line = line

        items: LStr = [s.strip() for s in line.split()]
        items = [s for s in items if s]
        self._flags = [s for s in items if s not in LOGS]
        self._logs = [s for s in items if s in LOGS]

    @property
    def logs(self) -> LStr:
        """ACE Option logs, items that not related to packet forwarding"""
        return self._logs

    # =========================== methods ============================

    def data(self, uuid: bool = False) -> DAny:
        """Returns *Option* data as *dict*
        :param uuid: Returns self.uuid in data
        :type uuid: bool

        :return: Option data

        :example:
        option = Option("ack log")
            option.data() ->
                {"line": "ack log",
                 "platform": "ios",
                 "note": "",
                 "flags": ["ack"],
                 "logs": ["log"]}
        """
        data = dict(
            # init
            line=self.line,
            platform=self._platform,
            note=self.note,
            # property
            flags=self._flags,
            logs=self._logs,
        )
        if uuid:
            data["uuid"] = self.uuid
        return data



"""CISCO config parser"""

import re
from abc import ABC
from copy import deepcopy

from cisco_acl import helpers as h
from cisco_acl.types_ import DAny, DLStr, DStr, LDAny, LStr, OLStr


class ConfigParser(ABC):
    """CISCO config parser"""

    def __init__(self, **kwargs):
        """CISCO config parser
        :param config: Cisco config, "show running-config" output
        :param platform: Platform: "ios", "nxos"
        :param version: Software version (not implemented, planned for compatability)
        """
        self.platform: str = h.init_platform(**kwargs)
        self.version: str = str(kwargs.get("version") or "")
        self.config: str = str(kwargs.get("config") or "")  # not parsed config

        self.lines: LStr = []  # config in *list* format
        self.dic: DLStr = {}  # config in *dict* format, commands as *LStr*
        self.mdic: DAny = {}  # config in multidimensional *dict* format, commands as *LStr*
        self.dic_text: DStr = {}  # config in *dict* format, commands as *str*
        self.mdic_text: DAny = {}  # config in multidimensional *dict* format, commands as *str*

    def __repr__(self):
        name = self.__class__.__name__
        platform = self.platform
        version = self.version
        return f"<{name}: {platform=} {version=}>"

    # =========================== methods ============================

    def addgrs(self) -> LDAny:
        """Parses address groups from config
        :return: *dict* ready for AddrGroup

        :example:
            config: "object-group ip address NAME
                       10 host 10.0.0.1
                       20 10.0.0.0/24"
            self.platform: "nxos"
            return: [{"name": "NAME",
                      "items": ["10 host 10.0.0.1", "20 10.0.0.0/24"],
                      "platform": "nxos"}]
        """
        addgrs: LDAny = []
        for objgr_key, objgr_cfg in self.dic_text.items():
            regex = "object-group (network |ip address )?(.+)"
            type_, name = h.findall2(regex, objgr_key)
            if type_ and name:
                items = h.lines_wo_spaces(objgr_cfg)
                addgr_d: DAny = dict(name=name, items=items, platform=self.platform)
                addgrs.append(addgr_d)
        return addgrs

    # noinspection PyShadowingBuiltins,PyIncorrectDocstring
    def acls(self, type: str = "", **kwargs) -> LDAny:  # pylint: disable=redefined-builtin
        """Parses ACLs from config

        :param type: ACL type: "extended", "standard", "any" (default "any")
        :type type: str

        :param names: Parse only ACLs with specified names
        :type names: List[str]

        :return: Parsed ACLs
        :rtype: List[dict]

        :example:
            self.config: "ip access-list ACL_NAME
                            10 remark ACE_NAME1
                            20 permit icmp any any
                            30 remark ACEG_NAME2
                            40 deny ip any any
                          interface Ethernet1/1/1
                            ip access-group ACL_NAME in
                          "
            self.platform: "nxos"
            return: [{"acl_name": "ACL_NAME",
                      "aces": "10 remark ACE_NAME1
                               20 permit icmp any any
                               30 remark ACEG_NAME2
                               40 deny ip any any",
                      "input": ["interface Ethernet1/1/1"],
                      "output": [],
                      "platform": "nxos"}]
        """
        names: OLStr = kwargs.get("names")
        if names is not None:
            names = [str(s) for s in names]

        acls: LDAny = []  # result
        for acl_key, acl_cfg in self.dic_text.items():
            regex = "ip access-list (extended |standard )?(.+)"
            acl_type, name = h.findall2(regex, acl_key)
            if not name:
                continue
            if names is None or name in names:
                acl_type = h.init_type(type=acl_type, platform=self.platform)
                acl_d: DAny = dict(line=f"{acl_key}\n{acl_cfg}",
                                   platform=self.platform,
                                   name=name,
                                   type=acl_type,
                                   input=[],
                                   output=[])
                if not type or type == acl_type:
                    acls.append(acl_d)
        self._add_acl_interfaces(acls)
        return acls

    def pattern__cfg_acl(self) -> str:
        """Pattern for extended ACL, by platform"""
        if self.platform == "nxos":
            return "ip access-list "
        return "ip access-list extended "

    def pattern__object_group(self) -> str:
        """Pattern for object-group, by platform"""
        if self.platform == "nxos":
            return "object-group network "
        return "object-group ip address "

    # ========================= parse_config =========================

    def parse_config(self) -> None:
        """Parses config rows to specific format: list, dict, multidimensional dict
        # make rows, main_rows, dic, mdic, etc.
        self.dic_text - config in *dict* format, commands as *str*
        self.mdic_text - config in multidimensional *dict* format, commands as *str*
        """
        config_l = [s.rstrip() for s in self.config.splitlines()]
        config_l = [s for s in config_l if not re.match(r"!|$", s)]
        if config_l:
            config_l[0] = config_l[0].strip()
        self.lines = self._parse_lines(config_l)
        self.dic = self._parse_dic(config_l)
        self.mdic = self._parse_mdic(config_l)
        self.dic_text = {k: "\n".join(v) for k, v in self.dic.items()}
        self.mdic_text = self._join_mdic_text(self.mdic)

    @staticmethod
    def _parse_lines(config_l: LStr) -> LStr:
        """Returns command lines without indentation"""
        lines = [s.strip() for s in config_l]
        lines = [s for s in lines if s]
        return lines

    @staticmethod
    def _join_mdic_text(mdic: DAny) -> DAny:
        """Joins self.mdic[key]["_config_"] from List[str] to str"""

        def join_config(mdic_text_: DAny) -> None:
            """join self.mdic[key]["_config_"] from List[str] to str"""
            for key, values in mdic_text_.items():
                if key == "_config_" and isinstance(values, list):
                    mdic_text_[key] = "\n".join(values)
                else:
                    join_config(values)

        mdic_text: DAny = deepcopy(mdic)
        join_config(mdic_text)
        return mdic_text

    @staticmethod
    def _parse_dic(config_l: LStr) -> DLStr:
        """Config in dictionary format (indented strings in dictionary)
        :example:
            data: {"interface Ethernet1/1": ["ip address 1.1.1.1/24",
                                              "no shutdown",
                                              "hsrp 5",
                                              "ip 1.1.1.2"]}
        """
        data: DLStr = {}
        key = ""
        for line in config_l:
            # not indented lines (main config) used as dictionary keys
            if re.match(r"\S", line):
                if re.match("interface .+", key) and not data.get(key):
                    data[key] = []  # empty dict for interface without settings
                key = line.strip()
            # indented lines used as dictionary values
            elif re.match(r"\s", line):
                # save line as dictionary value | init
                line_ = line.strip()
                data[key] = data[key] + [line_] if data.get(key) else [line_]
        return data

    def _parse_mdic(self, config_l: LStr) -> DAny:
        """Parses config in multidimensional dict format
        :example:
            data: {"interface Ethernet1/1": {"_config_": ["ip address 1.1.1.1/24",
                                                           "no shutdown"],
                                              "hsrp 5" : {"_config_": ["ip 1.1.1.2"]}}
                   }
        """
        data: DAny = {"_config_": []}  # result
        # validation
        if not config_l:  # exit if config is empty
            return data
        if re.match(r"\s", config_l[0]):  # error if first line is incorrect
            raise ValueError("first line in config should not be indented")
        config_l.append("END_OF_CONFIG")  # add last line, to detect end of config

        # Foreach config, indented lines add to dictionary as value.
        # More indented line add to more embedded dictionary (dictionary of dictionary)
        i_max = len(config_l)
        i_next = 0
        indent_i = ""  # no indentation
        for i in range(i_max):
            # last line, END_OF_CONFIG
            if i_max - i <= 1:
                break
            # skip already processed lines
            if i_next and i <= i_next:
                continue

            indent_next = re.sub(r"(^\s*).*", r"\1", config_l[i + 1])  # indentation in next line
            line_i = config_l[i]

            # not indented lines, main config
            if indent_i == indent_next:
                data["_config_"].append(line_i.strip())  # add not indented line to config section

            # next line is indented, make more deep indentation parsing
            elif indent_i < indent_next:
                # indented config parser
                (i_next, indent_next, mdic) = self._get_indented_dic(i, config_l)
                data.update(mdic)  # save

        # Interface can be without indented configuration, but should be in "mdic" as key
        for line in config_l:
            # Skip not Interface lines
            if not re.match("interface .+", line, re.I):
                continue
            line = line.strip()
            # Skip already created keys
            if data.get(line):
                continue
            # make key with empty config
            data.update({line: {"_config_": []}})

        return data

    # =========================== helpers ============================

    def _add_acl_interfaces(self, acls: LDAny) -> None:
        """Adds input/output interfaces to parsed `acls`
        :result: Side effect `acls`
        """
        intf_acls_all: LDAny = self._acls_on_interfaces()
        for acl_d in acls:
            intf_acls = [d for d in intf_acls_all if d["name"] == acl_d["name"]]
            for intf_acl in intf_acls:
                if intf_acl["input"]:
                    acl_d["input"].append(intf_acl["input"])
                if intf_acl["output"]:
                    acl_d["output"].append(intf_acl["output"])
        for acl_d in acls:
            acl_d["input"] = sorted(set(acl_d["input"]))
            acl_d["output"] = sorted(set(acl_d["output"]))

    def _acls_on_interfaces(self) -> LDAny:
        """Returns data of ACLs applied to the interfaces
        :example:
            self.config: "interface GigabitEthernet1/1/1
                            ip address 10.0.2.1 255.255.255.0
                            ip access-group ACL_NAME in"
            return: [{"acl": "ACL_NAME",
                      "input": "interface GigabitEthernet1/1/1",
                      "output": ""}]
        """
        access_groups: LDAny = []
        intfs_cfg: DStr = self._interfaces_w_acl()
        for intf_name, intf_cfg in intfs_cfg.items():
            if not intf_name.startswith("interface "):
                raise ValueError("invalid interface")
            if access_group_t := re.findall(r"ip access-group (\S+) (\S+)", intf_cfg):
                acl_name = access_group_t[0][0]
                data: DAny = dict(name=acl_name, input="", output="")
                for acl_name, direction in access_group_t:
                    if not acl_name:
                        raise ValueError(f"absent access-group {acl_name=}")
                    if direction not in ["in", "out"]:
                        raise ValueError(f"invalid access-group {direction=}")
                    if direction == "in":
                        data.update(dict(input=intf_name))
                    elif direction == "out":
                        data.update(dict(output=intf_name))
                access_groups.append(data)
        return access_groups

    def _get_indented_dic(self, i, config_l) -> tuple:
        """Config in multidimensional dict format,
        multi indented strings as dictionary in dictionary"""
        # init
        key = config_l[i].strip()
        i += 1
        i_next = 0
        data: DLStr = {"_config_": []}
        indent_i = re.sub(r"(^\s*)(.*)", r"\1", config_l[i])  # indentation in first line
        indent_next = ""

        # Foreach config, indented lines add to dictionary as value.
        # More indented line add to more embedded dictionary (dictionary of dictionary)
        i_max = len(config_l)

        # pylint: disable=redefined-argument-from-local
        for i in range(i, i_max):
            # skip already processed lines
            if i_next and i <= i_next:
                continue

            # end of config_l
            if i_max == i + 1:
                break

            # indentation in next line
            indent_next = re.sub(r"(^\s*).*", r"\1", config_l[i + 1])
            line_i = config_l[i]

            # not indented lines, main config
            if indent_i == indent_next:
                data["_config_"].append(line_i.strip())  # add not indented line to config section

            # next line is indented, make more deep indentation parsing
            elif len(indent_i) < len(indent_next):
                # indented config parser
                (i_next, indent_next, mdic) = self._get_indented_dic(i, config_l)
                data.update(mdic)  # save
                if len(indent_i) > len(indent_next):
                    i = i_next
                    break

            # end of indentation
            elif len(indent_i) > len(indent_next):
                data["_config_"].append(line_i.strip())  # add not indented line to config section
                break

        return i, indent_next, {key: data}

    def _interfaces_w_acl(self) -> DStr:
        """Returns dict of interfaces with access-group, skips interfaces without ACLs
        :example:
            self.config: "interface GigabitEthernet1/1/1
                            ip address 10.0.1.1 255.255.255.0
                            ip access-group ACL_NAME in
                          interface GigabitEthernet1/1/2
                            ip address 10.0.2.1 255.255.255.0"
            return: {"interface GigabitEthernet1/1/1":
                     "ip address 10.0.1.1 255.255.255.0\nip access-group ACL_NAME in"}
        """
        return {k: s for k, s in self.dic_text.items() if re.search("ip access-group", s, re.M)}



"""Address - Source or destination address in ACE"""
from __future__ import annotations

from functools import total_ordering
from typing import Dict, Iterable, List, Optional, Union

from cisco_acl import base_address
from cisco_acl.base_address import BaseAddress
from cisco_acl.types_ import LStr, DAny, LDAny


@total_ordering
class Address(BaseAddress):
    """Address - Source or destination address in ACE"""

    def __init__(self, line: str, **kwargs):
        """Address
        :param line: Address line
            Line pattern        Platform    Description
            ==================  ==========  ===========================
            A.B.C.D A.B.C.D     ios, nxos   Address and wildcard bits
            A.B.C.D/LEN         nxos        Network prefix
            any                 ios, nxos   Any host
            host A.B.C.D        ios         A single host
            object-group NAME   ios         Network object group
            addrgroup NAME      nxos        Network object group
        :type line: str

        :param platform: Platform: "ios", "nxos" (default "ios")
        :type platform: str

        Helpers
        :param note: Object description
        :type note: Any

        :param items: List of *Address* objects for address group
        :type items: str, List[str], dict, List[dict], Address, List[Address]

        :param max_ncwb: Max count of non-contiguous wildcard bits
        :type max_ncwb: int

        :example: wildcard
            address = Address("10.0.0.0 0.0.0.3", platform="ios")
            result:
                address.line == "10.0.0.0 0.0.0.3"
                address.addrgroup == ""
                address.ipnet == IPv4Network("10.0.0.0/30")
                address.prefix == "10.0.0.0/30"
                address.subnet == "10.0.0.0 255.255.255.252"
                address.wildcard == "10.0.0.0 0.0.0.3"

        :example: host
            address = Address("host 10.0.0.1", platform=="nxos")
            result:
                address.line == "10.0.0.1/32"
                address.addrgroup == ""
                address.ipnet == IPv4Network("10.0.0.1/32")
                address.prefix == "10.0.0.1/32"
                address.subnet == "10.0.0.1 255.255.255.255"
                address.wildcard == "10.0.0.1 0.0.0.0"

        :example: address group
            address = Address("object-group NAME", platform="ios")
            result:
                address.line == "object-group network NAME"
                address.addrgroup == "NAME"
                address.ipnet == None
                address.prefix == ""
                address.subnet == ""
                address.wildcard == ""
        """
        super().__init__(**kwargs)  # addrgroup, wildcard, items, max_ncwb, etc
        self._items: LAddress = []
        self.line = line
        if self._type == "addrgroup":
            self.items = kwargs.get("items") or []

    # =========================== property ===========================

    @property
    def items(self) -> LAddress:
        """List of *Address* objects for address group (type="addrgroup")"""
        return self._items

    @items.setter
    def items(self, items: LUAddress) -> None:
        items_ = self._init_items(items)
        self._items = [o for o in items_ if isinstance(o, Address)]

    # =========================== methods ============================


IAddress = Iterable[Address]
LAddress = List[Address]
OAddress = Optional[Address]
UAddress = Union[str, LStr, DAny, LDAny, Address, LAddress]
LUAddress = List[UAddress]
DLAddress = Dict[str, LAddress]
DDLAddress = Dict[str, DLAddress]


# ============================ functions =============================

def collapse(addresses: IAddress) -> LAddress:
    """Collapses a list of *Address* objects and deletes subnets in the shadow
        :param addresses: Iterable *Address* objects
        :return: List of collapsed *Address* objects

        :raises TypeError: Passed addresses not match conditions:
            - Item of `addresses` is not *Address*
            - Address is non-contiguous wildcard

        :example:
            wildcard = Address("10.0.0.0 0.0.0.1")
            host2 = Address("host 10.0.0.2")
            host3 = Address("host 10.0.0.3")
            collapse([wildcard, host2, host3]) -> [Address("10.0.0.0 0.0.0.3")]
    """
    addresses = list(addresses)
    for address in addresses:
        if not isinstance(address, Address):
            raise TypeError(f"{address=} {Address} expected")
    # noinspection PyProtectedMember
    collapsed = base_address.collapse_(addresses)
    collapsed_: LAddress = [o for o in collapsed if isinstance(o, Address)]
    return collapsed_



"""Remark - comments in ACL"""
from __future__ import annotations

from functools import total_ordering
from typing import List

from cisco_acl import helpers as h
from cisco_acl.base_ace import BaseAce
from cisco_acl.types_ import DAny


@total_ordering
class Remark(BaseAce):
    """Remark - comments in ACL"""

    def __init__(self, line: str = "", **kwargs):
        """Remark
        :param line: string of ACEs
        :type line: str

        :param platform: Platform: "ios", "nxos" (default "ios")
        :type platform: str

        Helpers
        :param note: Object description
        :type note: Any

        :example:
            line: "10 remark text"
            result:
                self.line: "10 remark text"
                self.sequence: 10
                self.action: "remark"
                self.text: "text"
        """
        self._action = "remark"
        self._sequence = 0
        self._text = ""
        super().__init__(**kwargs)
        if sequence := h.init_int(kwargs.get("sequence") or 0):
            self._sequence = sequence
        if kwargs.get("text"):
            self._text = h.init_remark_text(kwargs.get("text") or "")
        if line:
            self.line = line

    def __lt__(self, other) -> bool:
        """< less than"""
        if hasattr(other, "sequence"):
            if self._sequence == other.sequence:
                if isinstance(other, Remark):
                    return self._text < other.text
                return True
            return self._sequence < other.sequence
        if isinstance(other, str):
            return False
        return True

    # =========================== property ===========================

    @property
    def action(self) -> str:
        """ACE remark action

        :example:
            Remark("10 remark text")
            return: "remark"
        """
        return self._action

    @property
    def line(self) -> str:
        """ACE remark line

        :example:
            Remark("10 remark text")
            return: "10 remark text" """
        items = [self._sequence_s(), self._action, self._text]
        return " ".join([s for s in items if s])

    @line.setter
    def line(self, line) -> None:
        line = h.init_line(line)
        ace_d = h.parse_action(line)

        action = ace_d["action"]
        expected = "remark"
        if action != expected:
            raise ValueError(f"invalid {action=}, {expected=}")

        self._sequence = h.init_int(ace_d["sequence"])
        self._text = h.init_remark_text(ace_d["text"])

    @property
    def text(self) -> str:
        """ACE remark text

        :example:
            Remark("10 remark text")
            return: "text"
        """
        return self._text

    @text.setter
    def text(self, text: str) -> None:
        self._text = h.init_remark_text(text)

    # =========================== methods ============================

    def data(self, uuid: bool = False) -> DAny:
        """Converts *Remark* object to *dict*
        :param uuid: Returns self.uuid in data
        :type uuid: bool

        :return: Remark data
        """
        data = dict(
            # init
            line=self.line,
            platform=self._platform,
            note=self.note,
            # property
            sequence=self._sequence,
            action=self._action,
            text=self._text,
        )
        if uuid:
            data["uuid"] = self.uuid
        return data


LRemark = List[Remark]



"""ACL - Access Control List"""
from __future__ import annotations

from functools import total_ordering
from typing import Dict, Generator, List, Union

from cisco_acl import helpers as h
from cisco_acl.ace import Ace, LAce
from cisco_acl.ace_group import AceGroup, UAceg, UAce, LUAceg, OUAce, LUAce
from cisco_acl.remark import Remark
from cisco_acl.static import INDENTATION
from cisco_acl.types_ import LStr, UStr, DAny, DLStr, SStr, T2Str


@total_ordering
class Acl(AceGroup):
    """ACL - Access Control List"""

    def __init__(self, line: str = "", **kwargs):
        """ACL - Access Control List.
        This class implements most of the Python list methods: append(), extend(), sort(), etc.

        :param line: ACL config, "show running-config" output
        :type line: str

        :param platform: Platform: "ios", "nxos" (default "ios")
        :type platform: str

        :param input: Interfaces, where Acl is used on input
        :type input: str

        :param output: Interfaces, where Acl is used on output
        :type output: str

        Helpers
        :param note: Object description
        :type note: Any

        :param max_ncwb: Max count of non-contiguous wildcard bits
        :type max_ncwb: int

        :param indent: ACE lines indentation (default "  ")
        :type indent: str

        :param protocol_nr: Well-known ip protocols as numbers
            True  - all ip protocols as numbers
            False - well-known ip protocols as names (default)
        :type protocol_nr: bool

        :param port_nr: Well-known TCP/UDP ports as numbers
            True  - all tcp/udp ports as numbers
            False - well-known tcp/udp ports as names (default)
        :type port_nr: bool

        :param group_by: Startswith in remark line. ACEs group, starting from the Remark,
            where line startswith `group_by`, will be applied to the same AceGroup,
            until next Remark that also startswith `group_by`
        :type group_by: str

        Alternate way to get `name` and ACEs `items`, if `line` absent
        :param str type: ACL type: "extended", "standard" (default from `line`)

        :param name: ACL name (default from `line`)
        :type name: str

        :param items: ACEs items: *str*, *Ace*, *AceGroup*, *Remark* objects
            (default from `line`)
        :type items: List[Union[List[str], dict, List[Dict], Ace, Remark, AceGroup]]

        :example:
            acl = Acl(line="ip access-list extended NAME\nremark TEXT\npermit icmp any any"),
                      platform="ios",
                      input="interface FastEthernet1",
                      indent=" ")

            acl.line == "ip access-list extended NAME\n remark TEXT\n permit icmp any any"
            acl.platform == "ios"
            acl.name == "NAME"
            acl.items == [Remark("remark TEXT"), Ace("permit icmp any any")]
            acl.input == ["interface FastEthernet1"]
            acl.output == []
            acl.indent == " "
        """
        self._items: LUAceg = []  # type: ignore
        items = kwargs.get("items") or []
        if "items" in kwargs:
            del kwargs["items"]

        self._indent: str = self._init_indent(**kwargs)
        self.input: LStr = kwargs.get("input") or []
        self.output: LStr = kwargs.get("output") or []
        super().__init__(**kwargs)  # name, group_by, items, max_ncwb

        if name := str(kwargs.get("name") or ""):
            self.name = name
        if _type := str(kwargs.get("type") or ""):
            self._type = h.init_type(type=_type, platform=self._platform)
        if items:
            self.items = items
            return
        self.line = line

    def __hash__(self) -> int:
        return self.line.__hash__()

    def __eq__(self, other) -> bool:
        """== equality"""
        if self.__class__ == other.__class__:
            if self.__hash__() == other.__hash__():
                return True
        return False

    def __lt__(self, other) -> bool:
        """< less than"""
        if hasattr(other, "sequence"):
            if self._sequence == other.sequence:
                if isinstance(other, (Acl, Ace)):
                    return str(self) < str(other)
                return False
            return self._sequence < other.sequence
        return False

    def __repr__(self):
        params = self._repr__params()
        params = self._repr__add_param("input", params)
        params = self._repr__add_param("output", params)
        if self._indent != INDENTATION:
            params = self._repr__add_param("indent", params)
        params = self._repr__add_param("protocol_nr", params)
        params = self._repr__add_param("port_nr", params)
        kwargs = ", ".join(params)
        name = self.__class__.__name__
        return f"{name}({kwargs})"

    @staticmethod
    def _init_indent(**kwargs) -> str:
        """Init indentation"""
        indent = kwargs.get("indent")
        if indent is None:
            indent = INDENTATION
        return str(indent)

    # =========================== property ===========================

    @property
    def indent(self) -> str:
        """ACE lines indentation (default "  ")"""
        return self._indent

    @indent.setter
    def indent(self, indent: str) -> None:
        if indent is None:
            indent = INDENTATION
        self._indent = str(indent)

    @property
    def input(self) -> LStr:
        """Interfaces where Acl is used on input"""
        return self._input

    @input.setter
    def input(self, items: UStr) -> None:
        _items: LStr = h.convert_to_lstr(items=items)
        self._input = sorted(_items)

    @property  # type: ignore
    def items(self) -> LUAceg:  # type: ignore
        """List of ACE items: *Ace*, *Remark*, *AceGroup*"""
        return self._items

    @items.setter
    def items(self, items: LUAces) -> None:
        if isinstance(items, (str, dict, Ace, Remark, AceGroup)):
            items = [items]
        if not isinstance(items, (list, tuple, Generator)):
            raise TypeError(f"{items=} {list} expected")

        # items
        _items: LUAceg = []
        for item in items:
            # object
            if isinstance(item, (Ace, Remark, AceGroup)):
                item._platform = self._platform
                item._type = self._type
                _items.append(item)
            # dict
            elif isinstance(item, dict):
                aceg_o: UAceg = self._dict_to_aceg(**item)
                _items.append(aceg_o)
            # str
            elif isinstance(item, str):
                line = h.init_line(item)
                ace_o: UAce = self._line_to_ace(line)
                _items.append(ace_o)
            else:
                raise TypeError(f"{item=} {str} expected")
        self._items = _items

        if self._group_by:
            self.group(group_by=self._group_by)

    @property
    def line(self) -> str:
        """ACL config line"""
        items = []
        for item in self._items:
            if isinstance(item, AceGroup):
                for item_ in item.items:
                    items.append(item_)
                continue
            items.append(item)
        ace = "\n".join([f"{self._indent}{o}" for o in items])
        _line = "\n".join([self._cfg_acl_name(), ace])
        return _line

    @line.setter
    def line(self, line: str) -> None:
        items = h.lines_wo_spaces(line)
        if not items:
            return

        item1, *items = items
        acl_type, acl_name = self._parse_type_name(item1)
        self._type = acl_type
        self._name = acl_name

        aces: LUAceg = []
        for item in items:
            ace_o: OUAce = self._line_to_oace(line=item, warning=True)
            if isinstance(ace_o, (Ace, Remark)):
                aces.append(ace_o)
        self.items = aces

    @property
    def output(self) -> LStr:
        """Interfaces, where Acl is used on output"""
        return self._output

    @output.setter
    def output(self, items: UStr) -> None:
        _items: LStr = h.convert_to_lstr(items=items)
        self._output = sorted(_items)

    @property
    def platform(self) -> str:
        """Platform: "ios" Cisco IOS, "nxos" Cisco Nexus NX-OS"""
        return self._platform

    @platform.setter
    def platform(self, platform: str) -> None:
        platform = h.init_platform(platform=platform)
        if platform == "nxos":
            self.ungroup_ports()
        self._platform = platform
        for item in self._items:
            item.platform = self._platform

    # =========================== methods ============================

    def data(self, uuid: bool = False) -> DAny:
        """Converts *Acl* object to *dict*
        :param uuid: Returns self.uuid in data
        :type uuid: bool

        :return: data in *dict* format

        :example:
        acl = Acl("ip access-list extended NAME\n"
                  "  10 permit tcp host 10.0.0.1 10.0.0.0 0.0.0.3 eq www 443 log")
        acl.data -> {
            "line": "ip access-list extended NAME\n"
                    "  10 permit tcp host 10.0.0.1 10.0.0.0 0.0.0.3 eq www 443 log",
            "platform": "ios",
            "name": "NAME",
            "items": [
                {"line": "10 permit tcp host 10.0.0.1 10.0.0.0 0.0.0.3 eq www 443 log",
                 "platform": "ios",
                 "note": "",
                 "protocol_nr": False,
                 "port_nr": False,
                 "sequence": 10,
                 "action": "permit",
                 "protocol": {"line": "tcp",
                              "platform": "ios",
                              "note": "",
                              "protocol_nr": False,
                              "has_port": True,
                              "name": "tcp",
                              "number": 6},
                 "srcaddr": {"line": "host 10.0.0.1",
                             "platform": "ios",
                             "note": "",
                             "items": [],
                             "addrgroup": "",
                             "ipnet": IPv4Network("10.0.0.1/32"),
                             "prefix": "10.0.0.1/32",
                             "subnet": "10.0.0.1 255.255.255.255",
                             "wildcard": "10.0.0.1 0.0.0.0"},
                 "srcport": {"line": "",
                             "platform": "ios",
                             "note": "",
                             "protocol": "",
                             "port_nr": False,
                             "items": [],
                             "operator": "",
                             "ports": [],
                             "sport": ""},
                 "dstaddr": {"line": "10.0.0.0 0.0.0.3",
                             "platform": "ios",
                             "note": "",
                             "items": [],
                             "addrgroup": "",
                             "ipnet": IPv4Network("10.0.0.0/30"),
                             "prefix": "10.0.0.0/30",
                             "subnet": "10.0.0.0 255.255.255.252",
                             "wildcard": "10.0.0.0 0.0.0.3"},
                 "dstport": {"line": "eq www 443",
                             "platform": "ios",
                             "note": "",
                             "protocol": "tcp",
                             "port_nr": False,
                             "items": [80, 443],
                             "operator": "eq",
                             "ports": [80, 443],
                             "sport": "80,443"},
                 "option": "log"}
            ],
            "input": ["interface Ethernet1"],
            "output": ["interface Ethernet2"],
            "group_by": "",
            "note": "",
            "indent": "  ",
            "protocol_nr": False,
            "port_nr": False,
        }
        """
        data = dict(
            # init
            line=self.line,
            platform=self._platform,
            type=self._type,
            input=self._input.copy(),
            output=self._output.copy(),
            name=self._name,
            items=[o.data(uuid=uuid) for o in self._items],
            group_by=self._group_by,
            note=self.note,
            max_ncwb=self.max_ncwb,
            indent=self._indent,
            protocol_nr=self._protocol_nr,
            port_nr=self._port_nr,
        )
        if uuid:
            data["uuid"] = self.uuid
        return data

    def group(self, group_by: str) -> None:
        """Groups ACEs to *AceGroup* by `group_by` startswith in remarks
        :param group_by: Startswith in remark line. ACEs group, starting from the Remark,
            where line startswith `group_by`, will be applied to the same AceGroup,
            until next Remark that also startswith `group_by`
        :type group_by: str

        :example:
        group_by: "=== "
        self.items: [Remark("=== NAME1"),
                     Ace("permit tcp any any"),
                     Remark("=== NAME2"),
                     Ace("permit udp any any")]
        result:
         self.items: [AceGroup(items=[Remark("=== NAME1"), Ace("permit tcp any any")]),
                      AceGroup(items=[Remark("=== NAME2"), Ace("permit udp any any")])]
        """
        if not group_by:
            return
        ungrouped_l: LUAce = []
        for item in self._items:
            if isinstance(item, (Ace, Remark)):
                ungrouped_l.append(item)
            elif isinstance(item, AceGroup):
                _ungrouped = self._ungroup(item.items)
                ungrouped_l.extend(_ungrouped)

        grouped_items_d: Dict[str, LUAceg] = {}
        group_name = ""
        grouped_items_d[group_name] = []
        for item in ungrouped_l:
            if isinstance(item, Remark):
                if item.text.startswith(group_by):
                    group_name = item.text
                    grouped_items_d[group_name] = []
            grouped_items_d[group_name].append(item)

        grouped_items: LUAceg = []
        for group_name, aces_items in grouped_items_d.items():
            if aces_items:
                aceg_o = AceGroup(platform=self._platform,
                                  type=self._type,
                                  group_by=group_by,
                                  protocol_nr=self._protocol_nr,
                                  port_nr=self._port_nr,
                                  name=group_name,
                                  items=aces_items)
                grouped_items.append(aceg_o)
        self._items = grouped_items
        self._group_by = group_by

    def delete_shadow(self, skip: LStr = None) -> DLStr:
        """Removes ACEs in the shadow (in the bottom, without hits) from ACL
        :param skip: Skips checking specified address type: "addrgroup", "nc_wildcard"
        :return: *dict* Shading (in the top) and shadow (in the bottom) ACEs

        :example:
        acl = Acl("ip access-list extended NAME
                     permit ip 10.0.0.0 0.0.0.3 any
                     permit ip host 10.0.0.1 any
                     permit ip host 10.0.0.2 any
                     permit ip host 10.0.0.4 any")
        acl.delete_shadow() -> {"permit ip 10.0.0.0 0.0.0.3 any": ["permit ip host 10.0.0.1 any",
                                                                   "permit ip host 10.0.0.2 any"]}
        acl.line == "ip access-list extended NAME
                      permit ip 10.0.0.0 0.0.0.3 any
                      permit ip host 10.0.0.4 any"
        """
        shading_d: DLStr = self.shading(skip)
        if not shading_d:
            return {}
        shadow: LStr = [s for ls in shading_d.values() for s in ls]

        acl_new: Acl = self.copy()
        acl_new.ungroup()
        aces: LStr = [o.line for o in acl_new.items]

        shading_reversed = list(shading_d)
        shading_reversed.reverse()
        for top in shading_reversed:
            idx = aces.index(top) + 1
            items_top = acl_new.items[:idx]
            items_bot = acl_new.items[idx:]
            items_bot = [o for o in items_bot if o.line not in shadow]
            shadow = [s for s in shadow if top != s]
            acl_new.items = items_top + items_bot

        if self.group_by:
            acl_new.group(self.group_by)
        self.items = acl_new.items
        return shading_d

    def shadow_of(self, skip: LStr = None) -> LStr:
        """Returns ACEs in the shadow (in the bottom)
        NOTES:
        - Method compare *Ace* with the same action. ACEs where self.action=="permit" and
            other.action=="deny" not taken into account (skip checking)
        :param skip: Skips checking specified address type: "addrgroup", "nc_wildcard"
        :return: ACEs in the shadow

        :example:
        acl = Acl("ip access-list extended NAME
                     permit ip 10.0.0.0 0.0.0.3 any
                     permit ip host 10.0.0.1 any
                     permit ip host 10.0.0.2 any
                     permit ip host 10.0.0.4 any")
        acl.shadow_of() -> ["permit ip host 10.0.0.1 any", "permit ip host 10.0.0.2 any"]
        """
        shading_d: DLStr = self.shading(skip)
        shadow: LStr = [s for ls in shading_d.values() for s in ls]
        return shadow

    def shading(self, skip: LStr = None) -> DLStr:
        """Returns shading (in the top) and shadow (in the bottom) ACEs as *dict*,
        where *key* is shading rule, *value* shadow rules.
        NOTES:
        - Method compare *Ace* with the same action. ACEs where self.action=="permit" and
            other.action=="deny" not taken into account (skip checking)
        :param skip: Skips checking specified address type: "addrgroup", "nc_wildcard"

        :return: Shading (in the top) and shadow (in the bottom) ACEs
        :raises ValueError: addrgroup without addresses, non-contiguous wildcard
        :example:
        acl = Acl("ip access-list extended NAME
                     permit ip 10.0.0.0 0.0.0.3 any
                     permit ip host 10.0.0.1 any
                     permit ip host 10.0.0.2 any
                     permit ip host 10.0.0.4 any")
        acl.shading() -> {"permit ip 10.0.0.0 0.0.0.3 any": ["permit ip host 10.0.0.1 any",
                                                             "permit ip host 10.0.0.2 any"]}
        """
        acl_o = self.copy()
        acl_o.ungroup()
        aces = [o for o in acl_o.items if isinstance(o, Ace)]

        shading_d: DLStr = {}  # result
        shadow: SStr = set()
        for idx, ace_top in enumerate(aces):
            aces_bottom = aces[idx + 1:]
            for ace_bottom in aces_bottom:
                if ace_bottom.shadow_of(other=ace_top, skip=skip):
                    if ace_bottom.line not in shadow:
                        shading_d.setdefault(ace_top.line, []).append(ace_bottom.line)
                    shadow.add(ace_bottom.line)
        return shading_d

    def ungroup_ports(self) -> None:
        """Ungroups ACEs with multiple ports in single line ("eq" or "neq")
        to multiple lines with single port
        :example:
            acl = Acl("ip access-list extended NAME
                       permit tcp any eq 1 2 any eq 3 4")
            acl.split_ports()
            acl.line -> "ip access-list extended NAME
                           permit tcp any eq 1 any eq 3
                           permit tcp any eq 1 any eq 4
                           permit tcp any eq 2 any eq 3
                           permit tcp any eq 2 any eq 4"
        """
        _items: LUAceg = []
        for ace_o in self._items:
            if isinstance(ace_o, Ace):
                aces: LAce = ace_o.ungroup_ports()
                _items.extend(aces)
                continue
            if isinstance(ace_o, AceGroup):
                ace_o.ungroup_ports()
            _items.append(ace_o)
        self.items = _items
        if self._group_by:
            self.group(group_by=self._group_by)

    def ungroup(self) -> None:
        """Ungroups *AceGroup* to a flat list of *Ace* items
        :example:
        self.items: [Ace("permit icmp any any"),
                     AceGroup(items=[Ace("permit tcp any any"), Ace("permit udp any any")])]

         after acl.ungroup()
         self.items: [Ace("permit icmp any any"),
                      Ace("permit tcp any any"),
                      Ace("permit udp any any")]
        """
        self._group_by = ""
        self.items = list(self._ungroup(self._items))

    # =========================== helpers ============================

    def _cfg_acl_name(self) -> str:
        """Acl name line, with "ip access-list" keyword in beginning
        :return: Acl name line

        :example:
            self.name: "NAME"
            self.platform: "ios"
            return: "ip access-list extended NAME"

        :example:
            self.name: "NAME"
            self.platform: "nxos"
            return: "ip access-list NAME"
        """
        items = ["ip access-list"]
        if self._platform == "ios":
            items.append(self._type)
        items.append(self._name)
        return " ".join(items)

    def _parse_type_name(self, line: str) -> T2Str:
        """Parses ACL type and name from line "ip access-list " """
        expected = "ip access-list "
        if not line.startswith(expected):
            raise ValueError(f"{line=}, {expected=}")

        _line = h.findall1("ip access-list (.+)", line)
        if not _line:
            return "", ""

        _type = ""
        if self._platform == "nxos":
            _type = "extended"
            name = _line
        else:  # ios
            _type, name = h.findall2("(extended) (.+)", _line)
            if not _type:
                _type, name = h.findall2("(standard) (.+)", _line)
                if not _type:
                    name = _line
                    if name in ["extended", "standard"]:
                        raise ValueError(f"invalid {line=}")
        _type = h.init_type(type=_type, platform=self._platform)

        if name:
            h.check_name(name)
        return _type, name

    def _ungroup(self, items: list) -> Generator:
        """Ungroups AceGroup to a flat list of items"""
        for item in items:
            if isinstance(item, (Acl, AceGroup)):
                yield from self._ungroup(item.items)
            else:
                yield item


LAcl = List[Acl]
UAces = Union[str, LStr, dict, DAny, Generator, Ace, Remark, AceGroup]
LUAces = List[UAces]



"""Group of ACE (Access Control Entry).
These are multiple ACEe items, which must be in a certain order.
If you are changing *Ace* items order (sequence numbers) inside *Acl*,
the AceGroup behaves like a ACE item and order of ACE items inside AceGroup is not changed.
AceGroup is useful for freezing ACEs section, to hold "deny" after certain "permit".
"""
from __future__ import annotations

import logging
from functools import total_ordering
from typing import List, Optional, Union

from cisco_acl import helpers as h
from cisco_acl.ace import Ace, LAce
from cisco_acl.base_ace import BaseAce
from cisco_acl.group import Group
from cisco_acl.remark import Remark, LRemark
from cisco_acl.static import ACTIONS
from cisco_acl.types_ import DAny
from cisco_acl.wildcard import Wildcard

UAce = Union[Ace, Remark]
USAce = Union[Ace, Remark, str]

OUAce = Optional[UAce]
LUAce = List[UAce]
LUSAce = List[USAce]

ULAce = Union[LAce, LRemark, LUAce]


@total_ordering
class AceGroup(BaseAce, Group):
    """Group of ACE (Access Control Entry)"""

    def __init__(self, line: str = "", **kwargs):
        """Group of ACE (Access Control Entry)
        :param line: string of ACEs
        :type line: str

        :param platform: Platform: "ios", "nxos" (default "ios")
        :type platform: str

        Helpers
        :param note: Object description
        :type note: Any

        :param max_ncwb: Max count of non-contiguous wildcard bits
        :type max_ncwb: int

        :param protocol_nr: Well-known ip protocols as numbers
            True  - all ip protocols as numbers
            False - well-known ip protocols as names (default)
        :type protocol_nr: bool

        :param port_nr: Well-known TCP/UDP ports as numbers
            True  - all tcp/udp ports as numbers
            False - well-known tcp/udp ports as names (default)
        :type port_nr: bool

        :param group_by: Startswith in remark line. ACEs group, starting from the Remark,
            where line startswith `group_by`, will be applied to the same AceGroup,
            until next Remark that also startswith `group_by`
        :type group_by: str

        Alternate way to get `name` and ACEs `items`, if `line` absent
        :param str type: ACL type: "extended", "standard" (default "extended")

        :param name: Name of AceGroup, usually Remark.text of 1st self.items
        :type name: str

        :param items: An alternate way to create *AceGroup* object from a list of *Ace*
            (default from `line`)
        :type items: List[Union[Ace, Remark, str]]

        :example:
            aceg = AceGroup("10 permit icmp any any\n  20 deny ip any any")
            result:
                aceg.line == "10 permit icmp any any\n20 deny ip any any"
                aceg.platform == "ios"
                aceg.note == "description"
                aceg.sequence == 20  # Taking from the first ACE in items.
                aceg.items == [Ace("10 permit icmp any any"), Ace("20 deny ip any any")]
        """
        self._name = ""
        self._group_by = ""
        self._items: LUAce = []
        BaseAce.__init__(self, **kwargs)  # platform, type, note, sequence, protocol_nr, port_nr
        Group.__init__(self)
        if name := kwargs.get("name"):
            self._name = h.init_name(name)
        if group_by := str(kwargs.get("group_by") or ""):
            self._group_by = group_by
        # noinspection PyProtectedMember
        self.max_ncwb: int = Wildcard._init_max_ncwb(**kwargs)
        if items := kwargs.get("items") or []:
            self.items = items
            return
        self.line = line

    def __lt__(self, other) -> bool:
        """< less than"""
        if hasattr(other, "sequence"):
            if self._sequence == other.sequence:
                if other.__class__.__name__ == "Remark":
                    return False
                if isinstance(other, Ace):
                    return False
                if isinstance(other, AceGroup):
                    return str(self) < str(other)
                raise TypeError(f"{other=} {AceGroup} expected")
            return self._sequence < other.sequence
        return False

    # =========================== property ===========================

    @property
    def group_by(self) -> str:
        """Groups ACEs to *AceGroup* by startswith ot this value in remarks"""
        return self._group_by

    @property
    def items(self) -> LUAce:
        """List of ACE items: *Ace*, *Remark*, *AceGroup*"""
        return self._items

    @items.setter
    def items(self, items: LUSAce) -> None:
        if isinstance(items, (str, Ace, Remark)):
            items = [items]
        if not isinstance(items, (list, tuple)):
            raise TypeError(f"{items=} {list} expected")

        _items: LUAce = []  # result
        for item in items:
            if isinstance(item, (Ace, Remark)):
                item._platform = self._platform
                item._type = self._type
                _items.append(item)
            elif isinstance(item, dict):
                ace_o: UAce = self._dict_to_ace(**item)
                _items.append(ace_o)
            elif isinstance(item, str):
                line = h.init_line(item)
                item_ = self._line_to_ace(line)
                _items.append(item_)
            else:
                raise TypeError(f"{item=} {str} expected")
        self._items = _items

    @property
    def line(self) -> str:
        """Group of ACL config line"""
        lines = [o.line for o in self._items]
        return "\n".join(lines)

    @line.setter
    def line(self, line: str) -> None:
        lines = line.split("\n")
        lines = [h.init_line(s) for s in lines]
        items_ = [self._line_to_oace(s, warning=True) for s in lines]
        items: LUAce = [o for o in items_ if isinstance(o, (Ace, Remark))]
        self.items = items
        sequence = self._items[0].sequence if self._items else 0
        self.sequence = sequence

    @property
    def name(self) -> str:
        """Acl/AceGroup name"""
        return self._name

    @name.setter
    def name(self, name: str) -> None:
        """Acl/AceGroup name, without "ip access-list "
        Requirements:
        - length <= 100 chars
        - all chars are digits
        - first char is ascii_letters, other chars are ascii_letters and punctuation
        """
        self._name = h.init_name(name)

    @property
    def platform(self) -> str:
        """Platform
        - "ios" Cisco IOS (extended ACL)
        - "nxos" Cisco Nexus NX-OS
        """
        return self._platform

    @platform.setter
    def platform(self, platform: str) -> None:
        """Changes platform, normalizes self.items regarding the new platform
        :param platform: Platform: "ios", "nxos" (default "ios")
        """
        self._platform = h.init_platform(platform=platform)

        for item in self._items:
            item.type = self._type
            if self._platform == "nxos":
                self.ungroup_ports()
            item.platform = self._platform

        data = self.data(uuid=True)
        self.__init__(**data)  # type: ignore

    @property
    def type(self) -> str:
        """ACL type: standard, extended"""
        return self._type

    @type.setter
    def type(self, type_: str) -> None:
        type_ = h.init_type(type=type_, platform=self.platform)
        for item in self._items:
            item.type = type_
        self._type = type_

        data = self.data(uuid=True)
        self.__init__(**data)  # type: ignore

    # =========================== methods ============================

    def data(self, uuid: bool = False) -> DAny:
        """Converts *AceGroup* object to *dict*
        :param uuid: Returns self.uuid in data
        :type uuid: bool

        :return: data in *dict* format
        """
        data = dict(
            # init
            line=self.line,
            platform=self._platform,
            type=self._type,
            name=self._name,
            items=[o.data(uuid=uuid) for o in self._items],
            group_by=self._group_by,
            note=self.note,
            protocol_nr=self._protocol_nr,
            port_nr=self._port_nr,
            # property
            sequence=self._sequence,
        )
        if uuid:
            data["uuid"] = self.uuid
        return data

    def delete_note(self) -> None:
        """Deletes note in all children self.items: Ace, AceGroup, Remark"""
        self.note = ""
        for item in self._items:
            item.note = ""
            if isinstance(item, AceGroup):
                item.delete_note()

    def get_remark_name(self) -> str:
        """Returns Rule name from 1st remark"""
        remark = self.get_remark()
        name = h.parse_remark_name(text=remark.text, group_by=self.group_by)
        return name

    def get_remark(self) -> Remark:
        """Returns 1st remark with Rule name"""
        remark = self.items[0]
        if not isinstance(remark, Remark):
            raise TypeError(f"{remark=} {Remark} expected")
        return remark

    # noinspection PyIncorrectDocstring
    @h.check_start_step_sequence
    def resequence(self, start: int = 10, step: int = 10, **kwargs) -> int:
        """Resequence all AceGroup.items and change sequence numbers
        :param start: Starting sequence number. start=0 - delete all sequence numbers
        :param step: Step to increment the sequence number
        :param items: List of Ace objects (default self.items)
        :return: Last sequence number
        """
        items: LUAce = kwargs.get("items") or self._items
        sequence: int = int(start)
        count = len(items)

        for id_, item in enumerate(items, start=1):
            if isinstance(item, AceGroup):
                params = dict(items=item.items)
                sequence = self.resequence(start=sequence, step=step, **params)
            item.sequence = sequence
            if id_ < count:
                sequence += step
        return sequence

    def ungroup_ports(self) -> None:
        """Ungroups ACEs with multiple ports in single line ("eq" or "neq")
        to multiple lines with single port
        :example:
            aceg = AceGroup("permit tcp any eq 1 2 any eq 3 4")
            aceg.split_ports()
            aceg.line -> "permit tcp any eq 1 any eq 3
                          permit tcp any eq 1 any eq 4
                          permit tcp any eq 2 any eq 3
                          permit tcp any eq 2 any eq 4"
        """
        _items: LUAce = []
        for ace_o in self._items:
            if isinstance(ace_o, Ace):
                aces: LAce = ace_o.ungroup_ports()
                _items.extend(aces)
                continue
            _items.append(ace_o)
        self.items = _items

    # =========================== helpers ============================

    def _dict_to_ace(self, **kwargs) -> UAce:
        """Converts *dict* data to object: *Ace*, *Remark*
        :param kwargs: ACE data
        :return: ACE object: *Ace* or *Remark*

        :example: permit
            data: {line="permit ip any any"}
            return: Ace("permit ip any any")

        :example: remark
            data: {line="remark text"}
            return: Remark("text")
        """
        action = kwargs.get("action")
        if action not in ACTIONS:
            raise ValueError(f"invalid action in {kwargs=}, expected {ACTIONS=}")
        kwargs["platform"] = self._platform
        kwargs["type"] = self._type
        kwargs["protocol_nr"] = self._protocol_nr
        kwargs["port_nr"] = self._port_nr
        if action in ["remark"]:
            return Remark(**kwargs)
        return Ace(**kwargs)

    def _dict_to_aceg(self, **kwargs) -> UAceg:
        """Converts *dict* data to object: *AceGroup*, *Ace*, *Remark*
        :param kwargs: ACE data
        :return: ACE object: *AceGroup*, *Ace*, *Remark*"""
        items = kwargs.get("items")
        if isinstance(items, (list, tuple)):
            kwargs["platform"] = self._platform
            kwargs["type"] = self._type
            kwargs["protocol_nr"] = self._protocol_nr
            kwargs["port_nr"] = self._port_nr
            return AceGroup(**kwargs)
        return self._dict_to_ace(**kwargs)

    def _line_to_ace(self, line: str) -> UAce:
        """Converts config line to object: *Ace*, *Remark*
        :param line: ACE line
        :return: ACE object: *Ace* or *Remark*

        :example: permit
            line: "permit ip any any"
            return: Ace("permit ip any any")

        :example: remark
            line: "remark text"
            return: Remark("text")

        :example: not ACE line
            line: "text"
            return: None
        """
        action = h.parse_action(line)["action"]
        if action in ["remark"]:
            return Remark(line, platform=self._platform, type=self._type)
        ace_o = Ace(line=line,
                    platform=self._platform,
                    type=self._type,
                    protocol_nr=self._protocol_nr,
                    port_nr=self._port_nr,
                    max_ncwb=self.max_ncwb)
        return ace_o

    def _line_to_oace(self, line: str, warning: bool = False) -> OUAce:
        """Converts config line to object: *Ace*, *Remark*, None"""
        skip = ["", "statistics per-entry"]
        if line in skip:
            return None
        try:
            ace_o = self._line_to_ace(line)
        except ValueError:
            if warning:
                msg = f"{line=} does not match ACE pattern"
                logging.warning(msg)
            ace_o = None
        return ace_o


# ============================= helpers ==============================


LAceGroup = List[AceGroup]
UAceg = Union[Ace, Remark, AceGroup]
LUAceg = List[UAceg]



"""TCP/UDP ports and names mapping for Cisco ACL"""

from cisco_acl import helpers as h
from cisco_acl.types_ import DInt, DiStr, LStr

# cisco Nexus 3172T, NXOS: version 9.3(8)
# cisco ISR4331/K9, Cisco IOS XE Software, Version 16.09.06
TCP_NAME_PORT__BASE = {
    "bgp": 179,
    "chargen": 19,
    "cmd": 514,
    "daytime": 13,
    "discard": 9,
    "domain": 53,
    "echo": 7,
    "exec": 512,
    "finger": 79,
    "ftp": 21,
    "ftp-data": 20,
    "gopher": 70,
    "hostname": 101,
    "ident": 113,
    "irc": 194,
    "klogin": 543,
    "kshell": 544,
    "login": 513,
    "lpd": 515,
    "nntp": 119,
    "pim-auto-rp": 496,
    "pop2": 109,
    "pop3": 110,
    "smtp": 25,
    "sunrpc": 111,
    "tacacs": 49,
    "talk": 517,
    "telnet": 23,
    "time": 37,
    "uucp": 540,
    "whois": 43,
    "www": 80,
}

# cisco Nexus 3172T, NXOS: version 9.3(8)
TCP_NAME_PORT__NXOS = {
    "drip": 3949,
}
TCP_NAME_PORT__NXOS = {**TCP_NAME_PORT__BASE, **TCP_NAME_PORT__NXOS}

# cisco ISR4331/K9, Cisco IOS XE Software, Version 16.09.06
TCP_NAME_PORT__IOS = {
    "msrpc": 135,
    "onep-plain": 15001,
    "onep-tls": 15002,
    "syslog": 514,  # alias of cmd
}
TCP_NAME_PORT__IOS = {**TCP_NAME_PORT__BASE, **TCP_NAME_PORT__IOS}

# cisco Nexus 3172T, NXOS: version 9.3(8)
# cisco ISR4331/K9, Cisco IOS XE Software, Version 16.09.06
UDP_NAME_PORT__BASE = {
    "biff": 512,
    "bootpc": 68,
    "bootps": 67,
    "discard": 9,
    "dnsix": 195,
    "domain": 53,
    "echo": 7,
    "isakmp": 500,
    "mobile-ip": 434,
    "nameserver": 42,
    "netbios-dgm": 138,
    "netbios-ns": 137,
    "netbios-ss": 139,
    "non500-isakmp": 4500,
    "ntp": 123,
    "pim-auto-rp": 496,
    "rip": 520,
    "snmp": 161,
    "snmptrap": 162,
    "sunrpc": 111,
    "syslog": 514,
    "tacacs": 49,
    "talk": 517,
    "tftp": 69,
    "time": 37,
    "who": 513,
    "xdmcp": 177,
}
# cisco Nexus 3172T, NXOS: version 9.3(8)
UDP_NAME_PORT__NXOS = UDP_NAME_PORT__BASE

# cisco ISR4331/K9, Cisco IOS XE Software, Version 16.09.06
UDP_NAME_PORT__IOS = {
    "ripv6": 521,
}
UDP_NAME_PORT__IOS = {**UDP_NAME_PORT__BASE, **UDP_NAME_PORT__IOS}


class PortName:
    """TCP/UDP ports and names mapping for Cisco ACL"""

    def __init__(self, protocol: str = "tcp", **kwargs):
        """TCP/UDP ports and names
        :param protocol: Protocol: "tcp", "udp"
        :param platform: Platform: "ios", "nxos"
        :param version: Software version (not implemented, planned for compatability)
        """
        self.protocol = self._init_protocol(protocol)
        self.platform = h.init_platform(**kwargs)
        self.version = str(kwargs.get("version") or "").lower()

    def __repr__(self):
        name = self.__class__.__name__
        params = [
            f"protocol={self.protocol!r}",
            f"platform={self.platform!r}",
            f"version={self.version!r}",
        ]
        params_s = ", ".join(params)
        return f"{name}({params_s})"

    @staticmethod
    def _init_protocol(protocol: str) -> str:
        """Init protocol, converts tcp, udp numbers to names"""
        protocol = str(protocol).lower()
        expected = ["tcp", "udp"]
        if protocol in expected:
            return protocol
        if str(protocol) == "6":
            return "tcp"
        if str(protocol) == "17":
            return "udp"
        raise ValueError(f"invalid {protocol=}, {expected=}")

    def names(self) -> DInt:
        """Returns TCP/UDP names and ports based on Cisco platform and software version"""
        if self.protocol in ["tcp", "6"]:
            if self.platform == "ios":
                return TCP_NAME_PORT__IOS
            if self.platform == "nxos":
                return TCP_NAME_PORT__NXOS
            return {}
        if self.protocol in ["udp", "17"]:
            if self.platform == "ios":
                return UDP_NAME_PORT__IOS
            if self.platform == "nxos":
                return UDP_NAME_PORT__NXOS
            return {}
        return {}

    def ports(self) -> DiStr:
        """Returns TCP/UDP ports and names based on Cisco platform and software version"""
        if self.protocol in ["tcp", "6"]:
            if self.platform == "ios":
                name_port = {**TCP_NAME_PORT__BASE, **TCP_NAME_PORT__IOS}
                return self._swap(name_port)
            if self.platform == "nxos":
                name_port = {**TCP_NAME_PORT__BASE, **TCP_NAME_PORT__NXOS}
                return self._swap(name_port)
            return {}
        if self.protocol in ["udp", "17"]:
            if self.platform == "ios":
                name_port = {**UDP_NAME_PORT__BASE, **UDP_NAME_PORT__IOS}
                return self._swap(name_port)
            if self.platform == "nxos":
                name_port = {**UDP_NAME_PORT__BASE, **UDP_NAME_PORT__NXOS}
                return self._swap(name_port)
            return {}
        return {}

    @staticmethod
    def _swap(name_port: DInt) -> DiStr:
        """Swaps names and ports in dict"""
        data: DiStr = {}
        for name, port in name_port.items():
            if port not in data:
                data[port] = name
        return data


# ============================ functions =============================

def all_known_names() -> LStr:
    """Returns all known names, that can be used in Cisco ACL (platform does not matter)"""
    items = set()
    items.update(set(TCP_NAME_PORT__BASE))
    items.update(set(TCP_NAME_PORT__NXOS))
    items.update(set(TCP_NAME_PORT__IOS))
    items.update(set(UDP_NAME_PORT__BASE))
    items.update(set(UDP_NAME_PORT__NXOS))
    items.update(set(UDP_NAME_PORT__IOS))
    return sorted(items)



"""ACE IP protocol object"""
from __future__ import annotations

from functools import total_ordering
from typing import List

from cisco_acl import helpers as h
from cisco_acl.base import Base
from cisco_acl.static import NR_TO_PROTOCOL, ANY_PROTOCOLS
from cisco_acl.types_ import StrInt, DAny

PROTOCOL_IP = 0


@total_ordering
class Protocol(Base):
    """ACE IP protocol object"""

    def __init__(self, line: str = "", **kwargs):
        """ACE. IP protocol object
        :param line: IP protocol line
        :type line: str

        :param platform: Platform: "ios", "nxos" (default "ios")
        :type platform: str

        Helpers
        :param note: Object description
        :type note: Any

        :param protocol_nr: Well-known ip protocols as numbers
            True  - all ip protocols as numbers
            False - well-known ip protocols as names (default)
        :type protocol_nr: bool

        :param has_port: ACE has tcp/udp src/dst ports
            True  - ACL has tcp/udp src/dst ports
            False - ACL does not have tcp/udp src/dst ports (default)
        :type port_nr: bool

        :example:
            protocol = Protocol("tcp", platform="ios")
            result:
                protocol.line == "tcp"
                protocol.name == "tcp"
                protocol.number == 6
        :example:
        protocol = Protocol("255", platform="ios")
            result:
                protocol.line == "255"
                protocol.name == ""
                protocol.number == 255
        """
        self._number = PROTOCOL_IP
        self._protocol_nr = bool(kwargs.get("protocol_nr") or False)
        self._has_port = bool(kwargs.get("has_port") or False)
        super().__init__(**kwargs)
        self.line = line

    # ========================== redefined ===========================

    def __hash__(self) -> int:
        return self.line.__hash__()

    def __eq__(self, other) -> bool:
        """== equality"""
        return self.__hash__() == other.__hash__()

    def __lt__(self, other) -> bool:
        """< less than"""
        if self.__class__ == other.__class__:
            if self.number != other.number:
                return self.number < other.number
        return False

    def __repr__(self):
        params = super()._repr__params()
        params = self._repr__add_param("protocol_nr", params)
        params = self._repr__add_param("has_port", params)
        kwargs = ", ".join(params)
        name = self.__class__.__name__
        return f"{name}({kwargs})"

    # =========================== property ===========================

    @property
    def has_port(self) -> bool:
        """True if protocol is "tcp" or "udp" with specified port"""
        return self._has_port

    @has_port.setter
    def has_port(self, has_port: bool) -> None:
        self._has_port = bool(has_port)

    @property
    def line(self) -> str:
        """ACE protocol name: "ip", "icmp", "tcp", etc."""
        number = self._number
        if self._protocol_nr and not self._has_port:
            return str(number)
        if name := NR_TO_PROTOCOL[self._platform].get(number):
            return str(name)
        return str(number)

    @line.setter
    def line(self, line: str) -> None:
        line = h.int_to_str(line)
        if not line:
            self._number = PROTOCOL_IP
            return

        # permit 255 any any
        if line.isdigit():
            number: int = int(line)
            if not 0 <= number <= 255:
                raise ValueError(f"invalid protocol {number=}, expected 0..255")

        # permit ip any any
        else:
            number_ = ANY_PROTOCOLS.get(line)
            if number_ is None:
                raise ValueError(f"invalid protocol {line=}, expected={list(ANY_PROTOCOLS)}")
            number = int(number_)

        self._number = number

    @property
    def name(self) -> str:
        """ACE protocol name: "ip", "icmp", "tcp", etc."""
        number = self._number
        if name := NR_TO_PROTOCOL[self._platform].get(number):
            return str(name)
        return ""

    @name.setter
    def name(self, name: str) -> None:
        if not isinstance(name, str):
            raise TypeError(f"{name=} {str} expected")
        self.line = name

    @property
    def number(self) -> int:
        """ACE protocol number: 0..255, where 0="ip", 1="icmp", etc."""
        return self._number

    @number.setter
    def number(self, number: StrInt) -> None:
        self.line = h.init_number(number)

    @property
    def protocol_nr(self) -> bool:
        """Well-known ip protocols as numbers
            True  - all ip protocols as numbers
            False - well-known ip protocols as names (default)
        """
        return self._protocol_nr

    @protocol_nr.setter
    def protocol_nr(self, protocol_nr: bool) -> None:
        self._protocol_nr = bool(protocol_nr)

    # =========================== methods ============================

    def data(self, uuid: bool = False) -> DAny:
        """Converts *Protocol* object to *dict*
        :param uuid: Returns self.uuid in data
        :type uuid: bool

        :return: Protocol data

        :example:
            address = Protocol(line="tcp")
            address.data() -> {"line": "tcp",
                              "platform": "ios",
                              "note": "",
                              "protocol_nr": False,
                              "has_port": False,
                              "name": "tcp",
                              "number": 6}
        """
        data = dict(
            # init
            line=self.line,
            platform=self._platform,
            note=self.note,
            protocol_nr=self._protocol_nr,
            has_port=self._has_port,
            # property
            name=self.name,
            number=self._number,
        )
        if uuid:
            data["uuid"] = self.uuid
        return data


LProtocol = List[Protocol]



"""AddressAg - Address of AddrGroup. A "group-object" item of "object-group network " command"""
from __future__ import annotations

import re
from functools import total_ordering
from ipaddress import IPv4Network
from typing import Iterable, List, Optional, Union

from cisco_acl import base_address
from cisco_acl import helpers as h
from cisco_acl.base_address import BaseAddress
from cisco_acl.types_ import StrInt, LStr, DAny, LDAny
from cisco_acl.wildcard import Wildcard


@total_ordering
class AddressAg(BaseAddress):
    """AddressAg - Address of AddrGroup. A "group-object" item of "object-group network " command"""

    def __init__(self, line: str, **kwargs):
        """Address of AddrGroup
        :param line: Address line
            Line pattern        Platform    Description
            ==================  ==========  ================================
            description         ios         Address group description
            A.B.C.D A.B.C.D     ios         Network subnet and mask bits
            host A.B.C.D        ios, nxos   A single host
            group-object        ios         Nested address group name
            A.B.C.D A.B.C.D     nxos        Network subnet and wildcard bits
            A.B.C.D/LEN         nxos        Network prefix and length
        :type line: str

        :param platform: Platform: "ios", "nxos" (default "ios")
        :type platform: str

        Helpers
        :param note: Object description
        :type note: Any

        :param items: List of *AddressAg*
        :type items: str, List[str], dict, List[dict], AddressAg, List[AddressAg]

        :param max_ncwb: Max count of non-contiguous wildcard bits
        :type max_ncwb: int

        :example: wildcard
            address = AddressAg("10 10.0.0.0 0.0.0.3", platform="nxos")
            result:
                address.line == "10 10.0.0.0 0.0.0.3"
                address.addrgroup == ""
                address.ipnet == IPv4Network("10.0.0.0/30")
                address.prefix == "10.0.0.0/30"
                address.sequence == 10
                address.subnet == "10.0.0.0 255.255.255.252"
                address.wildcard == "10.0.0.0 0.0.0.3"

        :example: host
            address = AddressAg("host 10.0.0.1", platform="nxos")
            result:
                address.line == "10.0.0.1/32"
                address.addrgroup == ""
                address.ipnet == IPv4Network("10.0.0.1/32")
                address.prefix == "10.0.0.1/32"
                address.sequence == 0
                address.subnet == "10.0.0.1 255.255.255.255"
                address.wildcard == "10.0.0.1 0.0.0.0"

        :example: address group
            address = AddressAg("group-object NAME", platform="ios")
            result:
                address.line == "group-object NAME"
                address.addrgroup == "NAME"
                address.ipnet == None
                address.prefix == ""
                address.sequence == 0
                address.subnet == ""
                address.wildcard == ""
        """
        self._items: LAddressAg = []
        self._sequence: int = 0
        super().__init__(**kwargs)  # platform, note, line, addrgroup, ipnet, wildcard, items
        self.line = line
        if self._addrgroup:
            self.items = kwargs.get("items") or []

    # =========================== property ===========================

    @property
    def items(self) -> LAddressAg:
        """List of *AddressAg* objects for address group (type="addrgroup")"""
        return self._items

    @items.setter
    def items(self, items: LUAddressAg) -> None:
        items_ = self._init_items(items)
        self._items = [o for o in items_ if isinstance(o, AddressAg)]

    @property
    def line(self) -> str:
        """Address group address line

        :example:
            self: AddressAg("10 10.0.0.0/24", platform="nxos")
            return: "10 10.0.0.0/24"
        """
        line_ = super().line
        if self._sequence:
            return f"{self._sequence} {line_}"
        return line_

    @line.setter
    def line(self, line: str) -> None:
        line = h.init_line(line)
        line_d = h.parse_address(line)
        line = line_d["address"]
        self._sequence = h.init_int(line_d["sequence"])

        if self._is_address_host(line):
            self._line__host(line)
        elif self._is_address_prefix(line):
            self._line__prefix(line)
        elif self._is_address_wildcard(line):
            if self._platform == "nxos":
                self._line__wildcard(line)
            else:  # ios
                self._line__subnet(line)
        elif self._is_addrgroup(line):
            self._line_addrgroup(line)
        else:
            raise ValueError(f"invalid address {line=}")

    @property
    def platform(self) -> str:
        """Platform: "ios" Cisco IOS, "nxos" Cisco Nexus NX-OS"""
        return self._platform

    @platform.setter
    def platform(self, platform: str) -> None:
        """Changes platform, normalizes self regarding the new platform
        :param platform: Platform: "ios", "nxos" (default "ios")
        """
        line = self.line
        self._platform = h.init_platform(platform=platform)

        if self._is_addrgroup(line):
            self._type = "addrgroup"

        elif self._platform == "ios":
            if isinstance(self.ipnet, IPv4Network):
                if self.ipnet.prefixlen == 32:
                    self._type = "host"
                else:
                    self._type = "subnet"
            else:  # wildcard
                msg = f"non-contiguous wildcard={line!r} can not be transformed to subnet"
                raise ValueError(msg)
            self._sequence = 0

        elif self._platform == "nxos":
            if isinstance(self.ipnet, IPv4Network):
                self._type = "prefix"
            else:
                self._type = "wildcard"

        for item in self._items:
            item.platform = self._platform

        data = self.data(uuid=True)
        self.__init__(**data)  # type: ignore

    @property
    def sequence(self) -> int:
        """Address group address sequence number
        :return: Sequence number

        :example: Address with sequence number
            self: Address("111 10.0.0.0/24", platform="nxos")
            return: 111

        :example: Address without sequence number
            self: Address("10.0.0.0/24", platform="nxos")
            return: 0
        """
        return self._sequence

    @sequence.setter
    def sequence(self, sequence: StrInt) -> None:
        self._sequence = h.init_int(sequence)

    # =========================== methods ============================

    def data(self, uuid: bool = False) -> DAny:
        """Converts *AddressAg* object to *dict*
        :param uuid: Returns self.uuid in data
        :type uuid: bool

        :return: *AddressAg* data
        """
        data = super().data(uuid)
        data["sequence"] = self._sequence
        return data

    # =========================== helpers ============================

    def _cmd_addrgroup(self) -> str:
        """Address group line beginning
        :return: "group-object "
        """
        return "group-object "

    def _is_addrgroup(self, line: str) -> bool:
        """True if address is group "group-object NAME" """
        regex = "^group-object (.+)"
        return bool(re.match(regex, line))

    def _line_addrgroup(self, line):
        """Sets attributes for address group: "group-object NAME" """
        if self._platform == "nxos":
            raise ValueError(f"invalid address {line=} for platform={self._platform!r}")
        addrgroup = h.findall1("^group-object (.+)", line)
        h.check_name(addrgroup)
        self._type = "addrgroup"
        self._addrgroup = addrgroup
        self._wildcard = None

    def _line__prefix(self, line: str) -> None:
        """Sets attributes for prefix: A.B.C.D/LEN"""
        ipnet = h.prefix_to_ipnet(line)

        self._type = "prefix"
        if self._platform == "ios":
            subnet = ipnet.with_netmask.replace("/", " ")
            self._line__subnet(subnet)
            return

        self._addrgroup = ""
        wildcard = ipnet.with_hostmask.replace("/", " ")
        self._wildcard = Wildcard(wildcard, platform=self._platform, max_ncwb=self.max_ncwb)

    def _line__subnet(self, line: str) -> None:
        """Sets attributes for subnet "A.B.C.D A.B.C.D" """
        if line == "0.0.0.0 0.0.0.0" and self._platform == "ios":
            raise ValueError(f"{line!r} is denied for platform={self._platform!r}")

        self._type = "subnet"
        self._wildcard = Wildcard.fsubnet(line, platform=self._platform, max_ncwb=self.max_ncwb)
        self._addrgroup = ""


IAddressAg = Iterable[AddressAg]
LAddressAg = List[AddressAg]
OAddressAg = Optional[AddressAg]
UAddressAg = Union[str, LStr, DAny, LDAny, AddressAg, LAddressAg]
LUAddressAg = List[UAddressAg]
LUSAddressAg = List[Union[str, AddressAg]]


# ============================ functions =============================

def collapse(addresses: IAddressAg) -> LAddressAg:
    """Collapses a list of *AddressAg* objects and deletes subnets in the shadow
        :param addresses: Iterable *AddressAg* objects
        :return: List of collapsed *AddressAg* objects

        :raises TypeError: Passed addresses not match conditions:
            - Item of `addresses` is not *AddressAg*
            - AddressAg is non-contiguous wildcard

        :example:
            wildcard = AddressAg("10.0.0.0 255.255.255.254")
            host2 = AddressAg("host 10.0.0.2")
            host3 = AddressAg("host 10.0.0.3")
            collapse([wildcard, host2, host3]) -> [AddressAg("10.0.0.0 255.255.255.252")]
    """
    addresses = list(addresses)
    for address in addresses:
        if not isinstance(address, AddressAg):
            raise TypeError(f"{address=} {AddressAg} expected")
    # noinspection PyProtectedMember
    collapsed = base_address.collapse_(addresses)
    return [o for o in collapsed if isinstance(o, AddressAg)]



"""BaseAddress - Parent of: Address, AddressAg"""

import re
from abc import abstractmethod
from functools import total_ordering
from ipaddress import IPv4Network
from typing import Optional

from cisco_acl import helpers as h
from cisco_acl.base import Base
from cisco_acl.types_ import DAny, OIpNet, LIpNet, LStr
from cisco_acl.wildcard import Wildcard


@total_ordering  # type: ignore
class BaseAddress(Base):
    """BaseAddress - Parent of: Address, AddressAg"""

    def __init__(self, **kwargs):
        """BaseAddress
        :param platform: Platform: "ios", "nxos" (default "ios")
        :type platform: str

        Helpers
        :param note: Object description
        :type note: Any

        :param max_ncwb: Max count of non-contiguous wildcard bits
        :type max_ncwb: int
        """
        self._type = ""
        self._addrgroup: str = ""
        self._items = []
        self._wildcard = None
        super().__init__(**kwargs)  # platform, note
        # noinspection PyProtectedMember
        self.max_ncwb: int = Wildcard._init_max_ncwb(**kwargs)

    def __repr__(self):
        params = self._repr__params()
        params = self._repr__add_param("items", params)
        kwargs = ", ".join(params)
        name = self.__class__.__name__
        return f"{name}({kwargs})"

    # ========================== redefined ===========================

    def __hash__(self) -> int:
        wildcard = ""
        if isinstance(self._wildcard, Wildcard):
            wildcard = self._wildcard.line
        return (self._addrgroup, wildcard).__hash__()

    def __eq__(self, other) -> bool:
        """== equality"""
        if self.__class__ == other.__class__:
            return self.__hash__() == other.__hash__()
        return False

    def __lt__(self, other) -> bool:
        """< less than"""
        if self.__class__ == other.__class__:
            if self.ipnet and other.ipnet:
                return self.ipnet < other.ipnet
            if self.ipnet and not other.ipnet:
                return True
            if not self.ipnet and other.ipnet:
                return False
            if self.wildcard or other.wildcard:
                return self.wildcard < other.wildcard
            if self._addrgroup or other.addrgroup:
                return self._addrgroup < other.addrgroup
            return self.line < other.line
        return False

    def __contains__(self, other) -> bool:
        """Returns True if other.ipnet is subnet of self.ipnet"""
        self_ipnet = self._get_ipnet(self)
        if not self_ipnet:
            raise TypeError(f"{self_ipnet=} {IPv4Network} expected")

        # other=AddressAg
        if other_ipnet := self._get_ipnet(other):
            if other_ipnet.subnet_of(self_ipnet):
                return True
            return False

        # other=AddrGroup, other_items=LAddressAg
        if other_items := self._get_items(other):
            for other_item in other_items:
                other_ipnet = self._get_ipnet(other_item)
                if not other_ipnet:
                    raise TypeError(f"{other_ipnet=} {IPv4Network} expected")
                if other_ipnet.subnet_of(self_ipnet):
                    return True
            return False

        raise TypeError(f"{other=} type AddressAg expected")

    # =========================== property ===========================

    @property
    @abstractmethod
    def items(self):
        """List of *Address* or *AddressAg* objects for address group"""

    def _init_items(self, items) -> list:
        """Init items of *Address*, *AddressAg* objects for address group"""
        if isinstance(items, (str, dict, self.__class__)):
            items = [items]
        if not isinstance(items, (list, tuple)):
            raise TypeError(f"{items=} {list} expected")

        items_ = []
        for item in items:
            if isinstance(item, self.__class__):
                item.platform = self._platform
                items_.append(item)
            elif isinstance(item, dict):
                addr_o = self.__class__(**item)
                items_.append(addr_o)
            elif isinstance(item, str):
                line = h.init_line(item)
                addr_o = self.__class__(line=line, platform=self._platform, max_ncwb=self.max_ncwb)
                items_.append(addr_o)
            else:
                raise TypeError(f"{item=} {str} expected")
        return items_

    @property
    def addrgroup(self) -> str:
        """Address group name, if type="addrgroup". Value of "object-group NAME", "addrgroup NAME"
        :return: Address group name

        :example:
            self: Address("object-group NAME", platform="ios")
            return: "NAME"

        :example:
            self: Address("addrgroup NAME", platform="nxos")
            return: "NAME"
        """
        return self._addrgroup

    @property
    def ipnet(self) -> OIpNet:
        """Address IPv4Network object, None if type="addrgroup"
        :return: *IPv4Network* or None

        :example:
            self: Address("10.0.0.0 0.0.0.3", platform="ios")
            return: IPv4Network("10.0.0.0/30")

        :example:
            self: Address("object-group NAME", platform="ios")
            return: None
        """
        if not isinstance(self._wildcard, Wildcard):
            return None
        return self._wildcard.ipnet

    @property
    def line(self) -> str:
        """ACE address line: "A.B.C.D A.B.C.D", "A.B.C.D/LEN", "any", "host A.B.C.D",
        "object-group NAME", "addrgroup NAME"

        :example:
            self: Address("host 10.0.0.1", platform="nxos")
            return: "host 10.0.0.0.1"
        """
        if self._type == "addrgroup":
            return f"{self._cmd_addrgroup()}{self._addrgroup}"
        if self._type == "any":
            return "any"
        if self._type == "host":
            if not isinstance(self.ipnet, IPv4Network):
                raise TypeError(f"{self.ipnet=} {IPv4Network} expected")
            return f"host {self.ipnet.network_address}"
        if self._type == "prefix":
            return self.prefix
        if self._type == "subnet":
            return self.subnet
        return self.wildcard

    @line.setter
    def line(self, line: str) -> None:
        line = h.init_line(line)
        if self._is_address_any(line):
            self._line__any()
        elif self._is_address_host(line):
            self._line__host(line)
        elif self._is_address_prefix(line):
            self._line__prefix(line)
        elif self._is_address_wildcard(line):
            self._line__wildcard(line)
        elif self._is_addrgroup(line):
            self._line_addrgroup(line)
        else:
            raise ValueError(f"invalid address {line=}")

    @property
    def platform(self) -> str:
        """Platform: "ios" Cisco IOS, "nxos" Cisco Nexus NX-OS"""
        return self._platform

    @platform.setter
    def platform(self, platform: str) -> None:  # pylint: disable=too-many-branches)
        """Changes platform, normalizes self regarding the new platform
        :param platform: Platform: "ios", "nxos" (default "ios")
        """
        line = self.line
        self._platform = h.init_platform(platform=platform)

        if self._is_addrgroup(self.line) or self._is_addrgroup(line):
            self._type = "addrgroup"

        elif self._platform == "ios":
            if isinstance(self.ipnet, IPv4Network):
                if self.ipnet.prefixlen == 32:
                    self._type = "host"
                elif str(self.ipnet) == "0.0.0.0/0":
                    self._type = "any"
                else:
                    self._type = "wildcard"
            else:
                self._type = "wildcard"

        elif self._platform == "nxos":
            if isinstance(self.ipnet, IPv4Network):
                if str(self.ipnet) == "0.0.0.0/0":
                    self._type = "any"
                else:
                    self._type = "prefix"
            else:
                self._type = "wildcard"

        for item in self._items:
            item.platform = self._platform

        data = self.data(uuid=True)
        self.__init__(**data)  # type: ignore

    @property
    def prefix(self) -> str:
        """Address prefix, "" if type="addrgroup"
        :return: Subnet with prefix length "A.B.C.D/LEN"

        :example:
            self: Address("10.0.0.0 0.0.0.3", platform="ios")
            return: "10.0.0.0/32"
        """
        if not isinstance(self.ipnet, IPv4Network):
            return ""
        return str(self.ipnet)

    @prefix.setter
    def prefix(self, prefix: str) -> None:
        self.line = prefix

    @property
    def subnet(self) -> str:
        """Address subnet, "" if type="addrgroup"
        :return: Subnet with mask "A.B.C.D A.B.C.D"

        :example:
            self: Address("10.0.0.0 0.0.0.3", platform="ios")
            return: "10.0.0.0 255.255.255.252"
        """
        if not isinstance(self.ipnet, IPv4Network):
            return ""
        return self.ipnet.with_netmask.replace("/", " ")

    @property
    def wildcard(self) -> str:
        """Address wildcard, "" if type="addrgroup"
        :return: Subnet with wildcard mask "A.B.C.D A.B.C.D"

        :example:
            self: Address("10.0.0.0/30", platform="nxos")
            return: "10.0.0.0 0.0.0.3"
        """
        if not isinstance(self._wildcard, Wildcard):
            return ""
        return self._wildcard.line

    @property
    def type(self) -> str:
        """Address type: "addrgroup", "prefix", "subnet", "wildcard"
        :return: Address type

        :example:
            self: Address("10.0.0.0 0.0.0.3", platform="ios")
            return: "wildcard"
        """
        return self._type

    # =========================== methods ============================

    def data(self, uuid: bool = False) -> DAny:
        """Converts *Address* object to *dict*
        :param uuid: Returns self.uuid in data
        :type uuid: bool

        :return: Address data

        :example:
            address = Address("10.0.0.0/24", platform="nxos")
            address.data() ->
                {"line": "10.0.0.0/24",
                 "platform": "nxos",
                 "note": "",
                 "items": [],
                 "max_ncwb": 16,
                 "type": "prefix",
                 "addrgroup": "",
                 "ipnet": IPv4Network("10.0.0.0/24"),
                 "prefix": "10.0.0.0/24",
                 "subnet": "10.0.0.0 255.255.255.0",
                 "wildcard": "10.0.0.0 0.0.0.255"}
        """
        data = dict(
            # init
            line=self.line,
            platform=self._platform,
            note=self.note,
            items=[o.data(uuid=uuid) for o in self._items],
            max_ncwb=self.max_ncwb,
            # property
            type=self._type,
            addrgroup=self._addrgroup,
            ipnet=self.ipnet,
            prefix=self.prefix,
            subnet=self.subnet,
            wildcard=self.wildcard,
        )
        if uuid:
            data["uuid"] = self.uuid
        return data

    def ipnets(self) -> LIpNet:
        """All IPv4Networks, including address group and wildcard items
        :return: List of IPv4Network

        :example: contiguous wildcard
            Address("10.0.0.0 0.0.0.3").ipnets() -> [IPv4Network("10.0.0.0/30")]

        :example: non-contiguous wildcard
            self: Address("10.0.0.0 0.0.1.3")
            return: [IPv4Network("10.0.0.0/30"), IPv4Network("10.0.1.0/30")]
        """
        if isinstance(self.ipnet, IPv4Network):
            return [self.ipnet]

        if isinstance(self._wildcard, Wildcard):
            return self._wildcard.ipnets()

        ipnets: LIpNet = []
        if self.type == "addrgroup":
            for item in self._items:
                wildcard_o = getattr(item, "_wildcard")
                if not isinstance(wildcard_o, Wildcard):
                    raise TypeError(f"{self.line} {item=} {Wildcard} expected")
                # noinspection PyProtectedMember
                ipnets_ = item._wildcard.ipnets()
                ipnets.extend(ipnets_)
        return ipnets

    def prefixes(self) -> LStr:
        """All prefixes, including address group and wildcard items
        :return: Subnets with prefix length

        :example:
            self: Address("object-group NAME", platform="ios")
            return: ["10.0.1.0/30", "10.0.2.0/30"]

        :example:
            self: Address("10.0.0.0 0.0.0.3", platform="ios")
            return: ["10.0.0.0/30"]
        """
        ipnets = self.ipnets()
        return [str(o) for o in ipnets]

    def subnet_of(self, other) -> bool:
        """Checks self *Address* (all ipnets) is subnet of `other` *Address* (any of ipnet)
        :param other: Other *Address* (top)
        :type other: *Union[Address, AddressAg]*
        :return: True - if *Address* is subnet of `other` *Address*
        """
        tops = other.ipnets()
        bottoms = self.ipnets()
        is_subnet = h.subnet_of(tops=tops, bottoms=bottoms)
        return is_subnet

    def subnets(self) -> LStr:
        """All subnets, including address group and wildcard items
        :return: Subnets with mask

        :example:
            self: Address("object-group NAME", platform="ios")
            return: ["10.0.1.0 255.255.255.252", "10.0.2.0 255.255.255.252"]

        :example:
            self: Address("10.0.0.0 0.0.0.3", platform="ios")
            return: ["10.0.0.0 255.255.255.252"]
        """
        ipnets = self.ipnets()
        return [o.with_netmask.replace("/", " ") for o in ipnets]

    def wildcards(self) -> LStr:
        """All wildcards, including address group and wildcard items
        :return: LIst of wildcard *str*

        :example:
            self: Address("object-group NAME", platform="ios")
            return: ["10.0.1.0 0.0.0.3", "10.0.2.0 0.0.0.3"]

        :example:
            self: Address("10.0.0.0 0.0.0.3", platform="ios")
            return: ["10.0.0.0 0.0.0.3"]
        """
        if isinstance(self._wildcard, Wildcard):
            return [self._wildcard.line]
        wildcards: LStr = []
        for item in self._items:
            wildcard_o = getattr(item, "_wildcard")
            if not isinstance(wildcard_o, Wildcard):
                raise TypeError(f"{self.line} {item=} {Wildcard} expected")
            wildcards.append(wildcard_o.line)
        return wildcards

    # =========================== helpers ============================

    def _cmd_addrgroup(self) -> str:
        """Address group line beginning
        :return: "object-group " or "addrgroup "

        :example:
            self.platform: "ios"
            return: "object-group "

        :example:
            self.platform: "nxos"
            return: "addrgroup "
        """
        if self._platform == "nxos":
            return "addrgroup "
        return "object-group "

    @staticmethod
    def _get_ipnet(obj) -> OIpNet:
        """Gets IPv4Network from object"""
        if hasattr(obj, "ipnet"):
            if ipnet := getattr(obj, "ipnet"):
                if isinstance(ipnet, IPv4Network):
                    return ipnet
        return None

    @staticmethod
    def _get_items(obj) -> Optional[list]:
        """Gets list of items from object"""
        if hasattr(obj, "items"):
            if items := getattr(obj, "items"):
                if isinstance(items, list):
                    return items
        return None

    @staticmethod
    def _is_address_any(line: str) -> bool:
        """True if address is any"""
        return line == "any"

    def _is_addrgroup(self, line: str) -> bool:
        """True if address is group "object-group NAME" or "addrgroup NAME" """
        regex = f"^{self._cmd_addrgroup()}(.+)"
        return bool(re.match(regex, line))

    @staticmethod
    def _is_address_host(line: str) -> bool:
        """True if address is "host A.B.C.D" """
        regex = f"host {h.OCTETS}"
        return bool(re.match(regex, line))

    @staticmethod
    def _is_address_prefix(line: str) -> bool:
        """True if address is prefix "A.B.C.D/LEN" """
        regex = h.OCTETS + r"/\d+$"
        return bool(re.match(regex, line))

    @staticmethod
    def _is_address_subnet(line: str) -> bool:
        """True if address is subnet: "A.B.C.D A.B.C.D" """
        regex = f"{h.OCTETS} {h.OCTETS}"
        if re.match(regex, line):
            try:
                IPv4Network(line.replace(" ", "/"))
            except ValueError:
                return False
            return True
        return False

    @staticmethod
    def _is_address_wildcard(line: str) -> bool:
        """True if address is wildcard: "A.B.C.D A.B.C.D" """
        regex = f"{h.OCTETS} {h.OCTETS}"
        return bool(re.match(regex, line))

    def _line_addrgroup(self, line):
        """Sets attributes for address group: "object-group NAME" or "addrgroup NAME" """
        regex = f"^{self._cmd_addrgroup()}(.+)"
        addrgroup = h.findall1(regex, line)
        h.check_name(addrgroup)
        self._type = "addrgroup"
        self._addrgroup = addrgroup
        self._wildcard = None

    def _line__any(self) -> None:
        """ACE address line, any"""
        self._type = "any"
        self._addrgroup = ""
        wildcard = "0.0.0.0 255.255.255.255"
        self._wildcard = Wildcard(wildcard, platform=self._platform, max_ncwb=self.max_ncwb)

    def _line__host(self, line: str) -> None:
        """Sets attributes for host: host A.B.C.D"""
        ip_ = h.findall1(f"^host ({h.OCTETS})", line)
        self._type = "host"
        self._addrgroup = ""
        wildcard = f"{ip_} 0.0.0.0"
        self._wildcard = Wildcard(wildcard, platform=self._platform, max_ncwb=self.max_ncwb)

    def _line__prefix(self, line: str) -> None:
        """Sets attributes for prefix: A.B.C.D/LEN"""
        ipnet = h.prefix_to_ipnet(line)

        self._type = "prefix"
        if self.platform == "ios":
            self._type = "wildcard"

        self._addrgroup = ""
        wildcard = ipnet.with_hostmask.replace("/", " ")
        self._wildcard = Wildcard(wildcard, platform=self._platform, max_ncwb=self.max_ncwb)

    def _line__wildcard(self, line: str) -> None:
        """Sets attributes for wildcard: A.B.C.D A.B.C.D"""
        self._type = "wildcard"
        self._addrgroup = ""
        self._wildcard = Wildcard(line, platform=self._platform, max_ncwb=self.max_ncwb)


# ============================ functions =============================


def collapse_(addresses: list) -> list:
    """Collapses *LAddress*, *LAddressAg*
    :param addresses: List of Address objects
    :return: Collapsed Address objects
    :raises TypeError: Passed addresses not match: Address.ipnet is not *IPv4Network*
    """
    if not addresses:
        return []
    for address in addresses:
        if address.type == "wildcard" and not isinstance(address.ipnet, IPv4Network):
            raise TypeError(f"{address=} is non-contiguous wildcard")

    collapsed: LIpNet = []
    ipnets: LIpNet = [o for a in addresses for o in a.ipnets()]
    while ipnets:
        ipnet = ipnets.pop()
        if [o for o in ipnets if ipnet.subnet_of(o)]:
            continue
        supernet = ipnet.supernet()
        subnets = set(supernet.subnets())
        if subnets.issubset({*ipnets, ipnet}):
            if supernet not in ipnets:
                ipnets.insert(0, supernet)
            continue
        collapsed.append(ipnet)

    addresses_ = []  # result
    for ipnet_ in collapsed:
        addr_o = addresses[0].copy()
        addr_o.note = None
        addr_o.prefix = str(ipnet_)
        addresses_.append(addr_o)
    return sorted(addresses_)



"""ACE - Access Control Entry"""
from __future__ import annotations

from functools import total_ordering
from typing import List

from cisco_acl import helpers as h
from cisco_acl.address import Address
from cisco_acl.base_ace import BaseAce
from cisco_acl.option import Option
from cisco_acl.port import Port
from cisco_acl.protocol import Protocol
from cisco_acl.types_ import DAny, OBool, DStr, LStr


@total_ordering
class Ace(BaseAce):
    """ACE - Access Control Entry"""

    def __init__(self, line: str, **kwargs):
        """ACE - Access Control Entry
        :param line: ACE config, "show running-config" output
        :type line: str

        :param platform: Platform: "ios", "nxos" (default "ios")
        :type platform: str

        Helpers
        :param note: Object description
        :type note: Any

        :param max_ncwb: Max count of non-contiguous wildcard bits
        :type max_ncwb: int

        :param protocol_nr: Well-known ip protocols as numbers
            True  - all ip protocols as numbers
            False - well-known ip protocols as names (default)
        :type protocol_nr: bool

        :param port_nr: Well-known TCP/UDP ports as numbers
            True  - all tcp/udp ports as numbers
            False - well-known tcp/udp ports as names (default)
        :type port_nr: bool

        :example:
            ace=Ace("10 permit tcp host 10.0.0.1 eq 179 10.0.0.0 0.0.0.3 eq 80 443 log")

            ace.line == "10 permit tcp host 10.0.0.1 10.0.0.0 0.0.0.3 eq www 443 log"
            ace.platform == "ios"
            ace.sequence == 10
            ace.action == "permit"
            ace.protocol == Protocol("tcp")
            ace.srcaddr == Address("host 10.0.0.1")
            ace.srcport == Port("eq bgp")
            ace.dstaddr == Address("10.0.0.0 0.0.0.3")
            ace.dstport == Port("eq www 443")
            ace.option == Option("log")
        """
        self._action = ""
        self._protocol = Protocol()
        self._srcaddr = Address("any")
        self._srcport = Port()
        self._dstaddr = Address("any")
        self._dstport = Port()
        self._option = Option()
        super().__init__(**kwargs)  # platform, note, protocol_nr, port_nr, max_ncwb
        if srcaddr := kwargs.get("srcaddr") or {}:
            self._srcaddr = Address(**srcaddr)
        if dstaddr := kwargs.get("dstaddr") or {}:
            self._dstaddr = Address(**dstaddr)
        self.line = line

    # ========================== redefined ===========================

    # noinspection DuplicatedCode
    def __lt__(self, other) -> bool:
        """< less than"""
        if hasattr(other, "sequence"):
            # sequence
            if self._sequence != other.sequence:
                return self._sequence < other.sequence
            # object
            if other.__class__.__name__ == "Remark":
                return False
            if other.__class__.__name__ == "AceGroup":
                return True
            if isinstance(other, Ace):
                # protocol
                if self._protocol.number != other.protocol.number:
                    return self._protocol.number < other.protocol.number
                # srcaddr
                if self._srcaddr.ipnet != other.srcaddr.ipnet:
                    return self._lt__srcaddr(other)
                # srcport
                if self._srcport.operator and other.srcport.operator:
                    is_lt = self._lt__srcport(other)
                    if isinstance(is_lt, bool):
                        return is_lt
                # dstaddr
                if self._dstaddr.ipnet != other.dstaddr.ipnet:
                    return self._lt__dstaddr(other)
                # dstport
                if self._dstport.operator and other.dstport.operator:
                    is_lt = self._lt__dstport(other)
                    if isinstance(is_lt, bool):
                        return is_lt
                # option, addrgroup
                return self.line < other.line
        return False

    # =========================== property ===========================

    @property
    def action(self) -> str:
        """ACE action: "permit", "deny"
        :return: ACE action
        :example:
            self: Ace("10 permit ip any any")
            return: "permit"
        """
        return self._action

    @property
    def dstaddr(self) -> Address:
        """ACE source address: "any", "host A.B.C.D", "A.B.C.D A.B.C.D", "A.B.C.D/24",
            "object-group NAME"
        :return: ACE destination Address object

        :example: ios
            self: Ace("permit ip host 1.1.1.1 any")
            return: Address("host 1.1.1.1")

        :example: nxos
            self: Ace("10 permit ip host 1.1.1.1 any", platform="nxos")
            return: Address("1.1.1.1/32")
        """
        return self._dstaddr

    @property
    def dstport(self) -> Port:
        """ACE destination ports: "eq www 443", ""neq 1 2", "lt 2", "gt 2", "range 1 3"
        :return: ACE destination Port object

        :example:
            self: Ace("permit tcp host 1.1.1.1 eq www 443 any eq 1025 log")
            return: Port("eq 1025")
        """
        return self._dstport

    @property
    def line(self) -> str:
        """ACE config line
        :return: ACE config line

        :example:
            self: Ace("10 permit ip any any")
            return: "10 permit ip any any"
        """
        if self._type == "extended":
            items = [
                self._sequence_s(),
                self._action,
                self._protocol.line,
                self._srcaddr.line,
                self._srcport.line,
                self._dstaddr.line,
                self._dstport.line,
                self._option.line,
            ]
        else:  # standard
            items = [
                self._sequence_s(),
                self._action,
                self._srcaddr.line,
                self._option.line,
            ]
        return " ".join([s for s in items if s])

    @line.setter
    def line(self, line: str) -> None:
        line = h.init_line(line)

        # parse line to dict
        if ace_d := h.parse_ace_extended(line):
            self._type = "extended"
        else:
            if ace_d := h.parse_ace_standard(line):
                self._type = "standard"
            else:
                raise ValueError(f"invalid {line=}")
        self._check_parsed_elements(data=ace_d, line=line)

        self._sequence = h.init_int(ace_d["sequence"])
        self._action = h.init_ace_action(ace_d["action"])
        self._srcaddr = Address(ace_d["srcaddr"],
                                platform=self._platform,
                                items=self._srcaddr.items,
                                max_ncwb=self.max_ncwb)
        self._dstaddr = Address(ace_d["dstaddr"],
                                platform=self._platform,
                                items=self._dstaddr.items,
                                max_ncwb=self.max_ncwb)
        protocol_o = Protocol(line=ace_d["protocol"],
                              platform=self._platform,
                              port_nr=self._port_nr,
                              protocol_nr=self._protocol_nr)
        kwargs_port = dict(platform=self._platform, protocol=protocol_o.name, port_nr=self._port_nr)
        self._srcport = Port(ace_d["srcport"], **kwargs_port)
        self._dstport = Port(ace_d["dstport"], **kwargs_port)
        protocol_o.has_port = bool(self._srcport.line or self._dstport.line)
        self._protocol = protocol_o
        self._option = Option(ace_d["option"], platform=self._platform)

    @property
    def option(self) -> Option:
        """ACE option: "syn", "ack", "log", etc
        :return: ACE option
        :example:
            self: Ace("10 permit ip any any log")
            return: "log"
        """
        return self._option

    @property
    def platform(self) -> str:
        """Platform: "ios" Cisco IOS, "nxos" Cisco Nexus NX-OS"""
        return self._platform

    @platform.setter
    def platform(self, platform: str) -> None:
        """Changes platform, normalizes self.items regarding the new platform
        :param platform: Platform: "ios", "nxos" (default "ios")
        """
        self._platform = h.init_platform(platform=platform)
        self._protocol.platform = self._platform
        self._srcaddr.platform = self._platform
        self._srcport.platform = self._platform
        self._dstaddr.platform = self._platform
        self._dstport.platform = self._platform
        self._option.platform = self._platform
        data = self.data(uuid=True)
        self.__init__(**data)  # type: ignore

    @property
    def protocol(self) -> Protocol:
        """ACE protocol: "ip", "icmp", "tcp", etc.
        :return: ACE Protocol object

        :example:
            self: Ace("10 permit ip any any")
            return: Protocol("ip")
        """
        return self._protocol

    @property
    def srcaddr(self) -> Address:
        """ACE source address: "any", "host A.B.C.D", "A.B.C.D A.B.C.D", "A.B.C.D/24",
            "object-group NAME".
        :return: ACE source Address object

        :example: ios
            self: Ace("permit ip host 1.1.1.1 any")
            return: Address("host 1.1.1.1")

        :example: nxos
            self: Ace("10 permit ip host 1.1.1.1 any", platform="nxos")
            return: Address("1.1.1.1/32")
        """
        return self._srcaddr

    @property
    def srcport(self) -> Port:
        """ACE source ports: "eq www 443", ""neq 2", "lt 2", "gt 2", "range 1 3"
        :return: ACE source Port object

        :example:
            self: Ace("permit tcp host 1.1.1.1 eq www 443 any eq 1025 log")
            return: Port("eq www 443")
        """
        return self._srcport

    @property
    def type(self) -> str:
        """ACL type: standard, extended"""
        return self._type

    @type.setter
    def type(self, type_: str) -> None:
        type_ = h.init_type(type=type_, platform=self.platform)
        if self._type == "extended" and type_ == "standard":
            if self._srcaddr.addrgroup:
                addrgroup = self._srcaddr.addrgroup
                raise ValueError(f"mutually exclusive: type={type_!r}, {addrgroup=}")
            self._protocol.line = "ip"
            self._srcport.line = ""
            self._dstaddr.line = "any"
            self._dstport.line = ""
            self.option.line = ""
        self._type = type_
        data = self.data(uuid=True)
        self.__init__(**data)  # type: ignore

    # =========================== methods ============================

    def data(self, uuid: bool = False) -> DAny:
        """Converts *Ace* object to *dict*
        :param uuid: Returns self.uuid in data
        :type uuid: bool

        :return: ACE data

        :example:
            ace = Ace("10 permit tcp host 10.0.0.1 10.0.0.0 0.0.0.3 eq 80 443 log")
            ace.data() -> {
                "line": "10 permit tcp host 10.0.0.1 10.0.0.0 0.0.0.3 eq www log",
                "platform": "ios",
                "type": "extended",
                "sequence": 10,
                "action": "permit",
                "protocol": {"line": "tcp",
                             "platform": "ios",
                             "note": "",
                             "protocol_nr": False,
                             "has_port": True,
                             "name": "tcp",
                             "number": 6},
                "srcaddr": {"line": "host 10.0.0.1",
                            "platform": "ios",
                            "items": [],
                            "note": "",
                            "addrgroup": "",
                            "ipnet": IPv4Network("10.0.0.1/32"),
                            "prefix": "10.0.0.1/32",
                            "subnet": "10.0.0.1 255.255.255.255",
                            "wildcard": "10.0.0.1 0.0.0.0"},
                "srcport": {"line": "",
                            "platform": "ios",
                            "protocol": "",
                            "note": "",
                            "port_nr": False,
                            "items": [],
                            "operator": "",
                            "ports": [],
                            "sport": ""},
                "dstaddr": {"line": "10.0.0.0 0.0.0.3",
                            "platform": "ios",
                            "items": [],
                            "note": "",
                            "addrgroup": "",
                            "ipnet": IPv4Network("10.0.0.0/30"),
                            "prefix": "10.0.0.0/30",
                            "subnet": "10.0.0.0 255.255.255.252",
                            "wildcard": "10.0.0.0 0.0.0.3"},
                "dstport": {"line": "eq www 443",
                            "platform": "ios",
                            "protocol": "tcp",
                            "note": "",
                            "port_nr": False,
                            "items": [80,
                            443],
                            "operator": "eq",
                            "ports": [80,
                            443],
                            "sport": "80,443"},
                "option": {"line": "log",
                           "platform": "ios",
                           "note": "",
                           "flags": [],
                           "logs": ["log"]},
                "note": "a"}
        """
        data = dict(
            # init
            line=self.line,
            platform=self._platform,
            type=self._type,
            note=self.note,
            max_ncwb=self.max_ncwb,
            protocol_nr=self._protocol_nr,
            port_nr=self._port_nr,
            # property
            sequence=self._sequence,
            action=self._action,
            protocol=self._protocol.data(uuid=uuid),
            srcaddr=self._srcaddr.data(uuid=uuid),
            srcport=self._srcport.data(uuid=uuid),
            dstaddr=self._dstaddr.data(uuid=uuid),
            dstport=self._dstport.data(uuid=uuid),
            option=self._option.data(uuid=uuid),
        )
        if uuid:
            data["uuid"] = self.uuid
        return data

    def shadow_of(self, other: Ace, skip: LStr = None) -> bool:
        """Checks is ACE in the shadow of other ACE
        NOTES:
        - Method compare *Ace* with the same action. ACEs where self.action=="permit" and
            other.action=="deny" not taken into account (skip checking)
        :param other: Other *Ace* object (rule in the top)
        :param skip: Skips checking specified address type: "addrgroup", "nc_wildcard"

        :return: True - self *Ace* is in the shadow of other *Ace*
        :raises ValueError: addrgroup without addresses, non-contiguous wildcard

        :example: self is in the shadow of other, because tcp=2 is in tcp=[1, 2, 3]
            other.line == "permit tcp any any range 1 3"
            self.line == "permit tcp any any eq 2"
            return: True
        """
        if self._action != other.action:
            return False
        if not self._shadow_of__protocol(other):
            return False
        if not self._shadow_of__srcaddr(other=other, skip=skip):
            return False
        if not self._shadow_of__dstaddr(other=other, skip=skip):
            return False
        if not self._shadow_of__srcport(other):
            return False
        if not self._shadow_of__dstport(other):
            return False
        if not self._shadow_of__option(other):
            return False
        return True

    def ungroup_ports(self) -> LAce:
        """If self.srcport or self.dstport has "eq" or "neq" with multiple ports,
            then split them to multiple *Ace*
        :return: List of *Ace* with single port in each line
        :example:
            ace = Ace("permit tcp any eq 1 2 any eq 3 4", platform="ios")
            ace.split_ports -> [Ace("permit tcp any eq 1 any eq 3"),
                                Ace("permit tcp any eq 1 any eq 4"),
                                Ace("permit tcp any eq 2 any eq 3"),
                                Ace("permit tcp any eq 2 any eq 4")]
        """
        _aces: LAce = []
        if self.srcport.operator in ["eq", "neq"]:
            for item in self.srcport.items:
                ace_o = self.copy()
                ace_o.srcport.items = [item]
                _aces.append(ace_o)
        else:
            _aces.append(self.copy())

        aces = []  # result
        for ace_o_ in _aces:
            if ace_o_.dstport.operator in ["eq", "neq"]:
                for item in ace_o_.dstport.items:
                    ace_o = ace_o_.copy()
                    ace_o.dstport.items = [item]
                    aces.append(ace_o)
            else:
                aces.append(ace_o_.copy())
        if len(aces) == 1:
            return [self]
        return aces

    # =========================== helpers ============================

    @staticmethod
    def _check_parsed_elements(line: str, data: DStr) -> bool:
        """Checks parsed ACE elements
        :return: True if all elements are valid
        :raises: ValueError - some element in line is invalid
        """
        protocol = data["protocol"]
        srcport = data["srcport"]
        dstport = data["dstport"]
        if not (protocol or srcport or dstport):
            raise ValueError(f"invalid {line=}, required one of: {protocol=} {srcport=} {dstport=}")
        if protocol == "ip" and srcport:
            raise ValueError(f"invalid {line=}, mutually exclusive {protocol=} {srcport=}")
        if protocol == "ip" and dstport:
            raise ValueError(f"invalid {line=}, mutually exclusive {protocol=} {dstport=}")
        return True

    # noinspection DuplicatedCode
    def _shadow_of__srcaddr(self, other: Ace, skip: LStr = None) -> bool:
        """True if bottom address is in the shadow of the top address
        :param other: *Ace* in the top
        :param skip: Skips checking specified address type: "addrgroup", "nc_wildcard"
        """
        skip = skip or []
        if "addrgroup" in skip:
            if "addrgroup" in [self.srcaddr.type, other.srcaddr.type]:
                return False
        elif "nc_wildcard" in skip:
            if "wildcard" in [self.srcaddr.type, other.srcaddr.type]:
                if not (self.srcaddr.ipnet and other.srcaddr.ipnet):
                    return False

        tops = other.srcaddr.ipnets()
        bottoms = self._srcaddr.ipnets()
        is_subnet = h.subnet_of(tops=tops, bottoms=bottoms)
        return is_subnet

    # noinspection DuplicatedCode
    def _shadow_of__dstaddr(self, other: Ace, skip: LStr = None) -> bool:
        """True if bottom address is in the shadow of the top address
        :param other: *Ace* in the top
        :param skip: Skips checking specified address type: "addrgroup", "nc_wildcard"
        """
        skip = skip or []
        if "addrgroup" in skip:
            if "addrgroup" in [self.dstaddr.type, other.dstaddr.type]:
                return False
        elif "nc_wildcard" in skip:
            if "wildcard" in [self.dstaddr.type, other.dstaddr.type]:
                if not (self.dstaddr.ipnet and other.dstaddr.ipnet):
                    return False

        tops = other.dstaddr.ipnets()
        bottoms = self._dstaddr.ipnets()
        return h.subnet_of(tops=tops, bottoms=bottoms)

    def _shadow_of__protocol(self, other: Ace) -> bool:
        """True if self.protocol is in the shadow of the  other.protocol"""
        if other.protocol.name == "ip":
            return True
        return other.protocol.number == self._protocol.number

    def _shadow_of__srcport(self, other: Ace) -> bool:
        """True if self.srcport is in the shadow of the  other.srcport"""
        if top := set(other.srcport.ports):
            if bottom := set(self._srcport.ports):
                diff = bottom.intersection(top)
                return diff == bottom
            return False
        return True

    def _shadow_of__dstport(self, other: Ace) -> bool:
        """True if self.dstport is in the shadow of the  other.dstport"""
        if top := set(other.dstport.ports):
            if bottom := set(self._dstport.ports):
                diff = bottom.intersection(top)
                return diff == bottom
            return False
        return True

    def _shadow_of__option(self, other: Ace) -> bool:
        """True if self.dstport is in the shadow of the  other.dstport"""
        if top := set(other.option.flags):
            if bottom := set(self._option.flags):
                diff = bottom.intersection(top)
                return diff == bottom
            return False
        return True

    def _lt__srcaddr(self, other: Ace) -> bool:
        """< less than, srcaddr"""
        if self._srcaddr.ipnet and other.srcaddr.ipnet:
            return self._srcaddr.ipnet < other.srcaddr.ipnet
        if self._srcaddr.ipnet and not other.srcaddr.ipnet:
            return True
        return False

    def _lt__dstaddr(self, other: Ace) -> bool:
        """< less than, dstaddr"""
        if self._dstaddr.ipnet and other.dstaddr.ipnet:
            return self._dstaddr.ipnet < other.dstaddr.ipnet
        if self._dstaddr.ipnet and not other.dstaddr.ipnet:
            return True
        return False

    # noinspection DuplicatedCode
    def _lt__srcport(self, other: Ace) -> OBool:
        """< less than, srcport"""
        if self._srcport.operator != other.srcport.operator:
            return self._srcport.operator < other.srcport.operator
        if self._srcport.items[0] != other.srcport.items[0]:
            return self._srcport.items[0] < other.srcport.items[0]
        if self._srcport.items[-1] != other.srcport.items[-1]:
            return self._srcport.items[-1] < other.srcport.items[-1]
        return None

    # noinspection DuplicatedCode
    def _lt__dstport(self, other: Ace) -> OBool:
        """< less than, dstport"""
        if self._dstport.operator != other.dstport.operator:
            return self._dstport.operator < other.dstport.operator
        if self._dstport.items[0] != other.dstport.items[0]:
            return self._dstport.items[0] < other.dstport.items[0]
        if self._dstport.items[-1] != other.dstport.items[-1]:
            return self._dstport.items[-1] < other.dstport.items[-1]
        return None


LAce = List[Ace]
LLAce = List[LAce]



"""Base - Parent of: Address, Port, Protocol, BaseAce"""

from abc import ABC, abstractmethod
from typing import Any
from uuid import uuid1

from cisco_acl import helpers as h
from cisco_acl.static import IOS
from cisco_acl.types_ import LStr, DAny


class Base(ABC):
    """Base - Parent of: Address, Port, Protocol, BaseAce"""

    def __init__(self, **kwargs):
        """Base
        :param platform: Platform: "ios", "nxos" (default "ios")
        :type platform: str

        Helpers
        :param uuid: Unique identifier
        :type uuid: str

        :param note: Object description
        :type note: Any
        """
        self._platform: str = h.init_platform(**kwargs)
        self._uuid: str = self._init_uuid(**kwargs)
        self.note: Any = self._init_note(**kwargs)

    def __repr__(self):
        params = self._repr__params()
        kwargs = ", ".join(params)
        name = self.__class__.__name__
        return f"{name}({kwargs})"

    def __str__(self):
        return self.line

    @staticmethod
    def _init_uuid(**kwargs) -> str:
        """Init uuid"""
        if uuid := str(kwargs.get("uuid") or ""):
            return uuid
        return str(uuid1())

    @staticmethod
    def _init_note(**kwargs) -> Any:
        """Init note"""
        note = kwargs.get("note")
        if note is None:
            return ""
        return note

    # =========================== property ===========================

    @property
    def line(self) -> str:
        """Stub"""
        return ""

    @line.setter
    def line(self, line: str) -> None:  # pylint: disable=no-self-use
        """Stub"""
        return

    @property
    def platform(self) -> str:
        """Platform: "ios" Cisco IOS, "nxos" Cisco Nexus NX-OS"""
        return self._platform

    @platform.setter
    def platform(self, platform: str) -> None:
        """Changes platform
        :param platform: Platform: "ios", "nxos" (default "ios")
        """
        self._platform = h.init_platform(platform=platform)
        uuid = self.uuid
        self.line = self.line
        self.uuid = uuid

    @property
    def uuid(self) -> str:
        """Universally Unique Identifier"""
        return self._uuid

    @uuid.setter
    def uuid(self, uuid_: str) -> None:
        if not isinstance(uuid_, str):
            raise TypeError(f"{uuid_=} {str} expected")
        self._uuid = uuid_

    # =========================== methods ============================

    def copy(self):
        """Copies the self object"""
        kwargs = self.data()
        return self.__class__(**kwargs)

    @abstractmethod
    def data(self, uuid: bool = False) -> DAny:
        """Converts self object to *dict*
        :param uuid: Returns self.uuid in data
        :type uuid: bool

        :return: data in *dict* format
        """

    # =========================== helpers ============================

    def _repr__add_param(self, param: str, params: LStr) -> LStr:
        """Adds param to list of params"""
        if value := getattr(self, param):
            params.append(f"{param}={value!r}")
        return params

    def _repr__params(self) -> LStr:
        """Returns parameters for __repr__"""
        params: LStr = [f"{self.line!r}"]
        if self._platform != IOS:
            params.append(f"platform={self._platform!r}")
        params = self._repr__add_param("note", params)
        return params



"""ACE helper functions"""

import logging
import re
from functools import wraps
from ipaddress import IPv4Network
from string import ascii_letters, digits, punctuation
from time import time
from typing import Any, List, NamedTuple

from cisco_acl.port_name import all_known_names
from cisco_acl.static import ACTIONS, IOS, MAX_LINE_LENGTH, OPERATORS, PLATFORMS, SEQUENCE_MAX
from cisco_acl.types_ import DStr, LStr, StrInt, LInt, OInt, SInt, T2Str, T3Str, DInt, SStr, LIpNet

ALL_KNOWN_TUDP_NAMES = all_known_names()

OCTETS = r"\d+\.\d+\.\d+\.\d+"


# =============================== str ================================

def acl_help_to_name_port(output: str) -> DInt:
    """Transforms `output` of ACL help to *dict* where key is name, value is port
    :param output: ACL help
    :return: names and ports
    :example:
        output: "bgp          Border Gateway Protocol (179)"
        return: {"bgp": 179}
    """
    data: DInt = {}
    lines = [s.strip().replace("(", " ") for s in output.split("\n")]
    for line in lines:
        name, port = findall2(r"(\S+).+\s(\d+)\)$", line)
        if name and port:
            data[name] = int(port)
    return data


def findall1(pattern: str, string: str, flags=0) -> str:
    """Parses 1st item of re.findall(). If nothing is found, returns an empty string.
    Group with parentheses in pattern is required
    :return: Interested substring
    :example:
        pattern: "a(b)cde"
        string: "abcde"
        return: "b"
    """
    result = (re.findall(pattern=pattern, string=string, flags=flags) or [""])[0]
    if isinstance(result, str):
        return result
    if isinstance(result, tuple):
        return result[0]
    return ""


def findall2(pattern: str, string: str, flags=0) -> T2Str:
    """Parses 2 items of re.findall(). If nothing is found, returns 2 empty strings.
    Group with parentheses in pattern is required
    :return: Two interested substrings
    :example:
        pattern: "a(b)(c)de"
        string: "abcde"
        return: "b", "c"
    """
    result = (re.findall(pattern=pattern, string=string, flags=flags) or [("", "")])[0]
    if isinstance(result, tuple) and len(result) >= 2:
        return result[0], result[1]
    return "", ""


def findall3(pattern: str, string: str, flags=0) -> T3Str:
    """Parses 3 items of re.findall(). If nothing is found, returns 3 empty strings.
    Group with parentheses in pattern is required
    :return: Three interested substrings
    :example:
        pattern: "a(b)(c)(d)e"
        string: "abcde"
        return: "b", "c", "d"
    """
    result = (re.findall(pattern=pattern, string=string, flags=flags) or [("", "", "")])[0]
    if isinstance(result, tuple) and len(result) >= 3:
        return result[0], result[1], result[2]
    return "", "", ""


def check_line_length(line: str) -> bool:
    """True if line length <= 100 chars, else raise ERROR"""
    length = len(line)
    expected = MAX_LINE_LENGTH
    if length > expected:
        raise ValueError(f"invalid {line=} {length=}, {expected=}")
    return True


def check_name(name: str) -> bool:
    """True if first char is ascii_letters, other chars can be punctuation, else raise ERROR"""
    if not name:
        raise ValueError(f"absent {name=}")
    skip_chas = {"?"}
    valid_chars: SStr = set(ascii_letters + digits + punctuation).difference(skip_chas)
    if invalid_chars := list(set(name).difference(valid_chars)):
        raise ValueError(f"{invalid_chars=} in {name=}")
    return True


def init_ace_action(action: str = "permit") -> str:
    """Init ACE action: "permit", "deny" """
    action = str(action)
    expected = ["permit", "deny"]
    if action not in expected:
        raise ValueError(f"invalid {action=}, {expected=}")
    return action


def init_line(line: str) -> str:
    """Init line, replace spaces to one space, checks length <= 100 chars"""
    if not isinstance(line, str):
        raise TypeError(f"{line=} {str} expected")
    line = replace_spaces(line)
    return line


def init_name(name: str) -> str:
    """ACL or address group name, without "ip access-list "
    Requirements:
    - length <= 100 chars
    - all chars are digits
    - first char is ascii_letters, other chars are ascii_letters and punctuation
    """
    if not name:
        name = ""
    if not isinstance(name, str):
        raise TypeError(f"{name=} {str} expected")
    name = name.strip()
    check_line_length(name)
    return name


def init_number(number: StrInt) -> str:
    """Init number, convert *int* to *str*"""
    while True:
        if isinstance(number, int):
            if number < 0:
                raise ValueError(f"{number=} positive expected")
            break
        if isinstance(number, str):
            if number.isdigit():
                number = int(number)
                break
            raise ValueError(f"{number=} digit expected")
        raise TypeError(f"{number=} {str} {int} expected")
    return str(number)


# noinspection PyIncorrectDocstring
def init_platform(**kwargs) -> str:
    """Init device platform
    :param platform: Not checked platform: "cisco_ios", "cisco_nxos", "cnx", "ios", "nxos"
    :return: Valid platform: "ios", "nxos"
    """
    platform = kwargs.get("platform") or ""
    if not platform:
        platform = IOS
    if not isinstance(platform, str):
        raise TypeError(f"{platform=} {str} expected")
    if platform in ["cisco_ios"]:
        platform = "ios"
    if platform in ["cisco_nxos", "cnx"]:
        platform = "nxos"
    if platform not in PLATFORMS:
        raise ValueError(f"invalid {platform=}, expected={PLATFORMS}")
    return platform


# noinspection PyIncorrectDocstring
def init_protocol(line: str, **kwargs) -> str:
    """Init protocol, converts tcp, udp numbers to name
    :param line: Protocol line. Example "eq www"
    :param protocol: Protocol name or number. Example: "tcp"
    """
    if line:
        protocol = str(kwargs.get("protocol") or "")
        if protocol == "6":
            protocol = "tcp"
        if protocol == "17":
            protocol = "udp"
        expected = ["", "tcp", "udp"]
        if protocol in expected:
            return protocol
    return ""


def init_remark_text(text: str) -> str:
    """Init Remark.text
    :param text: text for remark
    """
    if not isinstance(text, str):
        raise TypeError(f"{text=} {str} expected")
    text = text.strip()
    if not text:
        raise ValueError(f"{text=} value required")
    return text


# noinspection PyIncorrectDocstring
def init_type(**kwargs) -> str:
    """Init ACL type: "extended", "standard"
    :param platform: Platform: "ios", "nxos" (default "ios")
    :param type: Not checked ACL type: "extended", "standard", "ip access-list extended", etc
    """
    platform = init_platform(**kwargs)
    _type = str(kwargs.get("type") or "").strip()
    if re.match("ip access-list extended ", _type):
        _type = "extended"
    if re.match("ip access-list standard ", _type):
        _type = "standard"
    if _type not in ["extended", "standard"]:
        _type = "standard"
        if platform == "nxos":
            _type = "extended"
    if platform == "nxos" and _type == "standard":
        expected = "extended"
        raise ValueError(f"invalid type={_type!r}, {expected=}")
    return _type


def int_to_str(line: StrInt) -> str:
    """Init line, *int* or *str* convert to *str*, replace spaces"""
    if isinstance(line, int):
        if line < 0:
            raise ValueError(f"{line=} positive expected")
        line = str(line)
    if not isinstance(line, str):
        raise TypeError(f"{line=} {str} expected")
    line = replace_spaces(line)
    return line


def lines_wo_spaces(line: str) -> LStr:
    """Splits line by "\n", replaces multiple white spaces with single space
    :line: Text joined by "\n"
    :return: Lines with single spaces and split by "\n"
    :example:
        line: "  10 deny icmp any any  \n  20 permit ip any any  "
        return: ["10 deny icmp any any", "20 permit ip any any"]
    """
    if not isinstance(line, str):
        raise TypeError(f"{line=} {str} expected")
    lines = [replace_spaces(s) for s in line.split("\n")]
    return [s for s in lines if s]


def parse_remark_name(text: str, group_by: str) -> str:
    """Parses Rule name from Remark.text
    :param text: Remark.text
    :param group_by: Acl.group_by
    :return: Rule name
    :example:
        self.group_by = "= "
        parse_aceg_name("= C-1, text", "= ") -> "C-1"
    """
    name = text.split(",")[0]
    name = name.strip().replace(group_by, "")
    return name


def re_find_t(regex: str, line: str) -> tuple:
    """Finds interested tuple
    :param regex: Regex pattern, should contain groups
    :param line: String body
    :return: Tuple with interested values or empty tuple
    """
    return (re.findall(regex, line) or [tuple()])[0]


def replace_spaces(line: str) -> str:
    """Replaces multiple white spaces with single space"""
    return " ".join(line.split())


# =============================== int ================================

def init_int(line: StrInt) -> int:
    """Converts *str* to positive *int*"""
    if not isinstance(line, (int, str)):
        raise TypeError(f"{line=} {int} or {str} expected")
    if isinstance(line, str):
        line = line.strip()
        line = line.split()[0] if line else "0"
        if not line.isdigit():
            raise ValueError(f"{line=}, digit expected")
    line_: int = int(line)
    if line_ < 0:
        raise ValueError(f"{line_=}, positive expected")
    return line_


# =============================== list ===============================

def convert_to_lstr(items: Any) -> LStr:
    """Converts items to *List[str]*. If items has other type, raise ERROR"""
    if not items:
        items = []
    if isinstance(items, str):
        items = [items]
    items = list(items)
    _items: LStr = []
    for item in items:
        if not isinstance(item, str):
            raise TypeError(f"{item=} {str} expected")
        _items.append(item)
    return _items


# =============================== dict ===============================

def parse_ace_extended(line: str) -> DStr:
    """Parses extended ACE line to *dict*
    :param line: ACE *str*
    :return: ACE *dict*

    :example:
        line: "10 permit tcp host 1.1.1.1 eq 1025 10.0.0.0 255.0.0.0 eq web log"
        return: {"sequence": "10",
                 "action": "permit",
                 "protocol": "tcp",
                 "srcaddr": "host 1.1.1.1",
                 "srcport": "eq 1025",
                 "dstaddr": "10.0.0.0 255.0.0.0",
                 "dstport": "eq web",
                 "option": "log"}
    """
    space = r"(?: )"
    text = r"\S+"
    addr = "|".join([
        "any",  # "any"
        f"host {OCTETS}",  # "host A.B.C.D"
        f"(?:object-group|addrgroup) {text}",  # ios: "object-group NAME", nxos: "addrgroup NAME"
        OCTETS + r"/\d+",  # "A.B.C.D/LEN"
        f"{OCTETS} {OCTETS}",  # "A.B.C.D A.B.C.D"
    ])

    re_sequence = r"(\d+)?"
    re_action = f"{space}?(permit|deny)"
    re_proto = f"({space}{text})?"
    re_srcaddr = f"{space}({addr})"
    re_srcport = "( .+)?"
    re_dstaddr = f"{space}({addr})"
    re_dstport = "( .+)?"

    regex = f"^{re_sequence}{re_action}{re_proto}{re_srcaddr}{re_srcport}{re_dstaddr}{re_dstport}"
    _items = re_find_t(regex, line)
    if not _items:
        return {}

    items = [s.strip() for s in _items]
    dstport_option = items[-1]
    result: DStr = _parse_dstport_option(dstport_option)
    data = dict(
        sequence=items[0],
        action=items[1],
        protocol=items[2],
        srcaddr=items[3],
        srcport=items[4],
        dstaddr=items[5],
        dstport=result["dstport"],
        option=result["option"],
    )
    return data


def parse_ace_standard(line: str) -> DStr:
    """Parses standard ACE line to *dict*
    :param line: ACE *str*
    :return: ACE *dict*

    :example:
        line: "10 permit host 1.1.1.1 log"
        return: {"sequence": "10",
                 "action": "permit",
                 "protocol": "ip",
                 "srcaddr": "host 1.1.1.1",
                 "srcport": "",
                 "dstaddr": "any",
                 "dstport": "",
                 "option": "log"}
    """
    space = r"(?: )"
    text = r"\S+"
    addr = "|".join([
        "any",  # "any"
        f"host {OCTETS}",  # "host A.B.C.D"
        f"(?:object-group|addrgroup) {text}",  # ios: "object-group NAME", nxos: "addrgroup NAME"
        OCTETS + r"/\d+",  # "A.B.C.D/LEN"
        f"{OCTETS} {OCTETS}",  # "A.B.C.D A.B.C.D"
    ])

    re_sequence = r"(\d+)?"
    re_action = f"{space}?(permit|deny)"
    re_srcaddr = f"{space}({addr})"
    re_log = "( .+)?"

    regex = f"^{re_sequence}{re_action}{re_srcaddr}{re_log}"
    _items = re_find_t(regex, line)
    if not _items:
        return {}

    items = [s.strip() for s in _items]
    data = dict(
        sequence=items[0],
        action=items[1],
        protocol="ip",
        srcaddr=items[2],
        srcport="",
        dstaddr="any",
        dstport="",
        option=items[3],
    )
    return data


def _parse_dstport_option(line: str) -> DStr:
    """Splits destination-ports and options based on end of ACE line
    :param line: ACE *str*
    :return: ACE *dict*
    :example:
        line: "eq bgp www ack log"
        return: {"dstport": "eq bgp www", "option": "ack log"}
    """
    dstports: LStr = []  # result
    options: LStr = []  # result
    if items := line.split():
        operator = items[0]
        if operator in OPERATORS:
            dstports.append(operator)
            items = items[1:]
            for id_, item in enumerate(items):
                if item.isdigit() or item in ALL_KNOWN_TUDP_NAMES:
                    dstports.append(item)
                    continue
                options = items[id_:]
                break
        else:
            options.extend(items)
    return dict(dstport=" ".join(dstports), option=" ".join(options))


def parse_action(line: str) -> DStr:
    """Parses action from ACE line

    :example:
        line: "10 remark TEXT"
        return: {"sequence": "10", "action": "remark", "text": "TEXT"}

    :example:
        line: "permit ip any any"
        return: "permit"
    """
    space = r"(?: )"
    actions = "|".join(ACTIONS)

    re_sequence = r"(\d+)?"
    re_action = f"{space}?({actions})"
    re_text = r"( .+)"
    regex = f"^{re_sequence}{re_action}{re_text}"
    _items = re_find_t(regex, line)
    if not _items:
        raise ValueError(f"invalid {line=}")
    items = [s.strip() for s in _items]
    data = dict(
        sequence=items[0],
        action=items[1],
        text=items[2],
    )
    return data


def parse_address(line: str) -> DStr:
    """Parses address item from address group
    :example:
        line: "10 host 10.0.0.1"
        return: {"sequence": 10, "address": "host 10.0.0.1"}
    """
    re_sequence = r"(\d+\s+)?"
    re_address = "(.+)"
    regex = f"^{re_sequence}{re_address}"
    _items = re_find_t(regex, line)
    if not _items:
        raise ValueError(f"invalid {line=}")
    items = [s.strip() for s in _items]
    data = dict(
        sequence=items[0],
        address=items[1],
    )
    return data


# ============================== ipnet ===============================

def prefix_to_ipnet(prefix: str) -> IPv4Network:
    """Converts prefix to ipnet, logging WARNING if invalid prefixlen
    :param prefix: Prefix "A.B.C.D/LEN"
    :return: ipnet *IPv4Network*
    """
    try:
        ipnet = IPv4Network(address=prefix)
    except ValueError as ex:
        if "has host bits set" not in str(ex):
            raise type(ex)(*ex.args)
        ipnet = IPv4Network(address=prefix, strict=False)
        msg = f"ValueError: {ex}, fixed to prefix {ipnet}"
        logging.warning(msg)
    return ipnet


def subnet_of(tops: LIpNet, bottoms: LIpNet) -> bool:
    """Checks all *IPv4Network*  in`bottoms` ara subnets of any *IPv4Network* in `tops`
    :param tops: List of *IPv4Network* in the top
    :type bottoms: List of *IPv4Network* in the bottom
    :return: True - if all `bottoms` atr subnets of `tops`
    """
    if not (tops and bottoms):
        return False
    for bottom in bottoms:
        for top in tops:
            if bottom.subnet_of(top):
                break
        else:
            return False
    return True


# ============================== ports ===============================

class PortRange(NamedTuple):
    """TCP/UDP ports range"""
    string: str
    range: range
    min: int
    max: int


def ports_to_string(items: LInt) -> str:
    """Converts list of ports to string
    :example:
        items: [1,3,4,5]
        return: "1,3-5"
    """
    if not items:
        return ""
    # if self.operator in ["eq", "neq"]:
    #     return ",".join([str(i) for i in items])

    items = sorted(items)
    ranges: LStr = []  # result
    item_1st: OInt = None
    for idx, item in enumerate(items, start=1):
        # not last iteration
        if idx < len(items):
            item_next = items[idx]
            if item_next - item <= 1:  # range
                if item_1st is None:  # start new range
                    item_1st = item
            else:  # int or end of range
                ranges.append(str(item) if item_1st is None else f"{item_1st}-{item}")
                item_1st = None
        # last iteration
        else:
            item_ = str(item) if item_1st is None else f"{item_1st}-{item}"
            ranges.append(item_)
    return ",".join(ranges)


def string_to_ports(ports: str) -> LInt:
    """Converts string to list of ports
    :example:
        ports: "1,3-5"
        return: [1, 3, 4, 5]
    """
    values = [s for s in ports.split(",") if s]
    ints: SInt = {int(s) for s in values if re.match(r"\d+$", s)}
    ranges = {s for s in values if re.match(r"\d+-\d+$", s)}
    ranges_t = _port_range_min_max(ranges)
    ports_calc = {i for o in ranges_t for i in o.range}  # ports in all ranges
    ports_calc.update(ints)
    ports_ = [i for i in ports_calc if 1 <= i <= 65535]
    return ports_


def _port_range_min_max(ranges) -> List[PortRange]:
    """Returns named tuple of range_sting, range_int, port_min, port_max"""
    ranges_tup = []
    for range_string in ranges:
        port_min, port_max = map(int, range_string.split("-"))
        range_int = range(port_min, port_max + 1)
        ranges_tup.append(PortRange(range_string, range_int, port_min, port_max))
    ranges_tup = sorted(ranges_tup, key=lambda o: o.min)
    return ranges_tup


# ============================= wrapper ==============================

def check_start_step_sequence(method):
    """Wrapper. Checks sequence numbers"""

    @wraps(method)
    def _wrapper(ace_o, start: int = 10, step: int = 10, **kwargs) -> int:
        """Checks sequence numbers, max/min value for `start` and `step`
        :param start: Starting sequence number. start=0 - delete all sequence numbers
        :param step: Step to increment the sequence number
        :param items: List of Ace objects (default self.items)
        :return: Last sequence number
        :raises: ValueError if `start` or `step` does not match conditions
        """
        if not 0 <= start <= SEQUENCE_MAX:
            raise ValueError(f"{start=} expected=0..{SEQUENCE_MAX}")
        if start and step < 1:
            raise ValueError(f"{step=} expected >= 1")
        if not start:
            step = 0

        sequence = method(ace_o, start, step, **kwargs)

        if sequence > SEQUENCE_MAX:
            raise ValueError(f"last {sequence=} expected=1..{SEQUENCE_MAX}")

        return sequence

    return _wrapper


def time_spent(func):
    """Wrapper measure function execution time"""

    def wrap(*args, **kwargs):
        """Wrapper"""
        started = time()
        _return = func(*args, **kwargs)
        elapsed = time() - started
        print("====== {:s}, spent {:.3f}s ======".format(func.__name__, elapsed))
        return _return

    return wrap



"""Typing"""
from ipaddress import IPv4Network, IPv4Address
from typing import Any, Dict, Iterable, List, Optional, Set, Tuple, Union

DAny = Dict[str, Any]
DInt = Dict[str, int]
DStr = Dict[str, str]
DiStr = Dict[int, str]
IInt = Iterable[int]
IStr = Iterable[str]
LAny = List[Any]
LBool = List[bool]
LInt = List[int]
LStr = List[str]
OBool = Optional[bool]
OInt = Optional[int]
OIpNet = Optional[IPv4Network]
OStr = Optional[str]
SInt = Set[int]
SStr = Set[str]
StrInt = Union[str, int]
T2IStr = Tuple[int, str]
T2Str = Tuple[str, str]
T3Str = Tuple[str, str, str]
TStr = Tuple[str, ...]
UIInt = Union[bytes, float, int, str, Iterable[Union[bytes, float, int, str]]]

DDAny = Dict[str, DAny]
DLAny = Dict[str, LAny]
DLStr = Dict[str, LStr]
LDAny = List[DAny]
LDStr = List[DStr]
LIpNet = List[IPv4Network]
LLStr = List[LStr]
LOIpNet = List[OIpNet]
LStrInt = List[StrInt]
LT2IStr = List[T2IStr]
OLStr = Optional[LStr]
T2IpAddr = Tuple[IPv4Address, IPv4Address]
TLintInt = Tuple[LInt, int]
UStr = Union[str, IStr]

DDLStr = Dict[str, DLStr]
DLDStr = Dict[str, LDStr]



"""Static settings"""

IOS = "ios"
MAX_LINE_LENGTH = 100
SEQUENCE_MAX = 4294967295
INDENTATION = "  "

PLATFORMS = ("ios", "nxos")
ACTIONS = ("remark", "permit", "deny")
OPERATORS = ("eq", "gt", "lt", "neq", "range")

OPTIONS = (
    "ack",
    "dscp",
    "fin",
    "log",
    "log-input",
    "match-all",
    "match-any",
    "precedence",
    "psh",
    "rst",
    "time-range",
    "tos",
    "ttl",
    "urg",
)
IOS_PROTOCOLS = {
    "ah": 51,
    "ahp": 51,
    "egp": 8,
    "eigrp": 88,
    "esp": 50,
    "gre": 47,
    "icmp": 1,
    "igmp": 2,
    "ip": 0,
    "ipip": 4,
    "ipv6": 41,
    "nos": 94,
    "ospf": 89,
    "pcp": 108,
    "pim": 103,
    "tcp": 6,
    "udp": 17,
}
NXOS_PROTOCOLS = {
    "ahp": 51,
    "eigrp": 88,
    "esp": 50,
    "gre": 47,
    "icmp": 1,
    "igmp": 2,
    "ip": 0,
    "nos": 94,
    "ospf": 89,
    "pcp": 108,
    "pim": 103,
    "tcp": 6,
    "udp": 17,
}
ANY_PROTOCOLS = {**IOS_PROTOCOLS, **NXOS_PROTOCOLS}
PROTOCOL_TO_NR = dict(
    ios=IOS_PROTOCOLS,
    nxos=NXOS_PROTOCOLS,
)
NR_TO_PROTOCOL = dict(
    ios={i: s for s, i in IOS_PROTOCOLS.items()},
    nxos={i: s for s, i in NXOS_PROTOCOLS.items()},
)



"""Package setup"""

import pathlib

from setuptools import setup  # type: ignore

import cisco_acl as package

VERSION = "2.1.0"
PACKAGE = package.__title__
PACKAGE_ = package.__title__.lower().replace("-", "_")  # PEP 503 normalization
ROOT = pathlib.Path(__file__).parent.resolve()
README = "README.rst"

if __name__ == "__main__":
    setup(
        name=PACKAGE_,
        packages=[PACKAGE_],
        package_data={PACKAGE_: ["py.typed"]},
        version=VERSION,
        description=package.__summary__,
        license=package.__license__,
        long_description=open(README).read(),
        long_description_content_type="text/x-rst",
        author=package.__author__,
        author_email=package.__email__,
        url=package.__url__,
        download_url=package.__download_url__,
        keywords="cisco, nexus, acl, ios, nx-os, networking, telecommunication",
        python_requires=">=3.8",
        install_requires=["netports"],
        classifiers=[
            # "Development Status :: 3 - Alpha",
            # "Development Status :: 4 - Beta",
            "Development Status :: 5 - Production/Stable",
            "Intended Audience :: Developers",
            "Intended Audience :: System Administrators",
            "Intended Audience :: Telecommunications Industry",
            # "Operating System :: Cisco IOS",
            # "Operating System :: Cisco NX-OS",
            "Topic :: System :: Networking",
            "License :: OSI Approved :: MIT License",
            "Programming Language :: Python :: 3.8",
            "Programming Language :: Python :: 3.9",
            "Programming Language :: Python :: 3.10",
            "Natural Language :: English",
        ],
    )



"""Group of Items"""

from __future__ import annotations

from typing import Iterable


class Group:
    """Group of Items"""

    def __init__(self, items=None):
        self.items = []
        if items:
            self.items = list(items)

    # ======================= special methods ========================

    def __add__(self, other) -> Group:
        """Returns self.items + other.items"""
        group = Group(self.items.copy())
        group.update(other.items)
        return group

    def __contains__(self, item) -> bool:
        """Returns key in self"""
        return item in self.items

    def __delitem__(self, idx: int) -> None:
        """Deletes self.items[key]"""
        self.items.__delitem__(idx)

    def __getitem__(self, idx: int):
        return self.items[idx]

    def __iter__(self):
        return iter(self.items)

    def __len__(self) -> int:
        return len(self.items)

    def __reversed__(self):
        """Returns a reverse iterator over the list"""
        for item in self.items[::-1]:
            yield item

    # =========================== methods ============================

    def add(self, item) -> None:
        """Adds new item to self.items list, if it is not in self.items"""
        if item not in self.items:
            self.items.append(item)

    def append(self, item) -> None:
        """Appends item to the end of the self.items list"""
        self.items.append(item)

    def clear(self) -> None:
        """Removes all items from the self.items list"""
        self.items = []

    def copy(self):
        """Returns a shallow copy of the self.items list"""
        return self.items.copy()

    def count(self, item):
        """Returns number of occurrences of the self.items"""
        return self.items.count(item)

    def delete(self, item) -> None:
        """Removes item from the self.items list"""
        if item in self.items:
            self.items.remove(item)

    def extend(self, items: Iterable) -> None:
        """Extends the self.items list by appending items"""
        if isinstance(items, (list, set, tuple)):
            self.items.extend(list(items))
            return
        raise TypeError(f"{items=} {list} expected")

    def index(self, *args) -> int:
        """Returns first index of item
        Raises ValueError if the value is not present"""
        return self.items.index(*args)

    def insert(self, *args) -> None:
        """Inserts item before index"""
        return self.items.insert(*args)

    def pop(self, *args):
        """Removes and return item at index (default last)
        Raises IndexError if list is empty or index is out of range"""
        return self.items.pop(*args)

    def remove(self, *args) -> None:
        """Removes first occurrence of items in the self.items
        Raises ValueError if the item is not present"""
        self.items.remove(*args)

    def reverse(self) -> None:
        """Reverses order of items in the self.items list"""
        self.items.reverse()

    def sort(self, *args, **kwargs) -> None:
        """Sorts the self.items list in ascending order
        :example:
            self.items.sort(reverse=True|False, key=myFunc)
        """
        self.items.sort(*args, **kwargs)

    def update(self, items: list) -> None:
        """Extends list by adding items to self.items list, if it is not in the self.items"""
        for item in items:
            self.add(item)



"""cisco-acl"""

from cisco_acl.ace import Ace
from cisco_acl.ace_group import AceGroup
from cisco_acl.acl import Acl
from cisco_acl.addr_group import AddrGroup
from cisco_acl.address import Address
from cisco_acl.address_ag import AddressAg
from cisco_acl.config_parser import ConfigParser
from cisco_acl.functions import aces, acls, addrgroups, range_ports, range_protocols
from cisco_acl.option import Option
from cisco_acl.port import Port
from cisco_acl.protocol import Protocol
from cisco_acl.remark import Remark
from cisco_acl.wildcard import Wildcard

__all__ = [
    "Ace",
    "AceGroup",
    "Acl",
    "AddrGroup",
    "Address",
    "AddressAg",
    "ConfigParser",
    "Option",
    "Port",
    "Protocol",
    "Remark",
    "Wildcard",
    "aces",
    "acls",
    "addrgroups",
]

__version__ = "2.1.0"
__date__ = "2022-10-25"
__title__ = "cisco-acl"

__summary__ = "Python package to parse and manage Cisco ACL (Access Control List)"
__author__ = "Vladimir Prusakov"
__email__ = "vladimir.prusakovs@gmail.com"
__url__ = "https://github.com/vladimirs-git/cisco-acl"
__download_url__ = f"{__url__}/archive/refs/tags/{__version__}.tar.gz"
__license__ = "MIT"
